# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import datetime
import platform
import re

import aiohttp
import mock
import pytest

from hikari import _about
from hikari import errors
from hikari import intents
from hikari import presences
from hikari import urls
from hikari.impl import config
from hikari.impl import shard
from hikari.internal import aio
from hikari.internal import net
from hikari.internal import time
from hikari.internal import ux
from tests.hikari import hikari_test_helpers


def test_log_filterer():
    filterer = shard._log_filterer(b"TOKEN")

    returned = filterer(b"this log contains the TOKEN and it should get removed and the TOKEN here too")
    assert returned == (
        b"this log contains the **REDACTED TOKEN** and it should get removed and the **REDACTED TOKEN** here too"
    )


def test__serialize_activity_when_activity_is_None():
    assert shard._serialize_activity(None) is None


def test__serialize_activity_when_activity_is_not_None():
    activity = presences.Activity(name="some name", type=0, state="blah", url="https://some.url")
    assert shard._serialize_activity(activity) == {
        "name": "some name",
        "type": 0,
        "state": "blah",
        "url": "https://some.url",
    }


@pytest.mark.parametrize(
    ("activity_name", "activity_state", "expected_name", "expected_state"),
    [("Testing!", None, "Custom Status", "Testing!"), ("Blah name!", "Testing!", "Blah name!", "Testing!")],
)
def test__serialize_activity_custom_activity_syntactic_sugar(
    activity_name, activity_state, expected_name, expected_state
):
    activity = presences.Activity(name=activity_name, state=activity_state, type=presences.ActivityType.CUSTOM)

    assert shard._serialize_activity(activity) == {
        "type": 4,
        "name": expected_name,
        "state": expected_state,
        "url": None,
    }


def test__serialize_datetime_when_datetime_is_None():
    assert shard._serialize_datetime(None) is None


def test__serialize_datetime_when_datetime_is_not_None():
    dt = datetime.datetime(2004, 11, 22, tzinfo=datetime.timezone.utc)
    assert shard._serialize_datetime(dt) == 1101081600000


@pytest.fixture
def http_settings():
    return mock.Mock(spec_set=config.HTTPSettings)


@pytest.fixture
def proxy_settings():
    return mock.Mock(spec_set=config.ProxySettings)


class StubResponse:
    def __init__(self, *, type=None, data=None, extra=None):
        self.type = type
        self.data = data
        self.extra = extra


class TestGatewayTransport:
    @pytest.fixture
    def transport_impl(self):
        return shard._GatewayTransport(
            ws=mock.Mock(),
            exit_stack=mock.AsyncMock(),
            logger=mock.Mock(),
            log_filterer=mock.Mock(),
            loads=mock.Mock(return_value={}),
            dumps=mock.Mock(),
            transport_compression=True,
        )

    def test_init_when_transport_compression(self):
        transport = shard._GatewayTransport(
            ws=mock.Mock(),
            exit_stack=mock.AsyncMock(),
            logger=mock.Mock(),
            log_filterer=mock.Mock(),
            loads=mock.Mock(),
            dumps=mock.Mock(),
            transport_compression=True,
        )

        assert transport._receive_and_check == transport._receive_and_check_zlib

    def test_init_when_no_transport_compression(self):
        transport = shard._GatewayTransport(
            ws=mock.Mock(),
            exit_stack=mock.AsyncMock(),
            logger=mock.Mock(isEnabledFor=mock.Mock(return_value=False)),
            log_filterer=mock.Mock(),
            loads=mock.Mock(),
            dumps=mock.Mock(),
            transport_compression=False,
        )

        assert transport._receive_and_check == transport._receive_and_check_text

    @pytest.mark.asyncio
    async def test_send_close(self, transport_impl):
        transport_impl._sent_close = False

        with mock.patch.object(asyncio, "wait_for", return_value=mock.AsyncMock()) as wait_for:
            with mock.patch.object(asyncio, "sleep") as sleep:
                await transport_impl.send_close(code=1234, message=b"some message")

        wait_for.assert_awaited_once_with(transport_impl._ws.close.return_value, timeout=5)
        transport_impl._ws.close.assert_called_once_with(code=1234, message=b"some message")
        transport_impl._exit_stack.aclose.assert_awaited_once_with()
        sleep.assert_awaited_once_with(0.25)

    @pytest.mark.asyncio
    async def test_send_close_when_TimeoutError(self, transport_impl):
        transport_impl._sent_close = False
        transport_impl._ws.close.side_effect = asyncio.TimeoutError

        with mock.patch.object(asyncio, "sleep") as sleep:
            await transport_impl.send_close(code=1234, message=b"some message")

        transport_impl._ws.close.assert_called_once_with(code=1234, message=b"some message")
        transport_impl._exit_stack.aclose.assert_awaited_once_with()
        sleep.assert_awaited_once_with(0.25)

    @pytest.mark.asyncio
    async def test_send_close_when_already_sent(self, transport_impl):
        transport_impl._sent_close = True

        with mock.patch.object(aiohttp.ClientWebSocketResponse, "close", side_effect=asyncio.TimeoutError) as close:
            await transport_impl.send_close(code=1234, message=b"some message")

        close.assert_not_called()

    @pytest.mark.asyncio
    @pytest.mark.parametrize("trace", [True, False])
    async def test_receive_json(self, transport_impl, trace):
        transport_impl._receive_and_check = mock.AsyncMock()
        transport_impl._logger = mock.Mock(enabled_for=mock.Mock(return_value=trace))

        assert await transport_impl.receive_json() == transport_impl._loads.return_value

        transport_impl._receive_and_check.assert_awaited_once_with()
        transport_impl._loads.assert_called_once_with(transport_impl._receive_and_check.return_value)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("trace", [True, False])
    async def test_send_json(self, transport_impl, trace):
        transport_impl._ws.send_bytes = mock.AsyncMock()
        transport_impl._logger = mock.Mock(enabled_for=mock.Mock(return_value=trace))
        transport_impl._dumps = mock.Mock(return_value=b"some data")

        await transport_impl.send_json({"json_send": None})

        transport_impl._ws.send_bytes.assert_awaited_once_with(b"some data")

    @pytest.mark.asyncio
    async def test__handle_other_message_when_TEXT(self, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.TEXT)

        with pytest.raises(errors.GatewayError, match="Unexpected message type received TEXT, expected BINARY"):
            transport_impl._handle_other_message(stub_response)

    @pytest.mark.asyncio
    async def test__handle_other_message_when_BINARY(self, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.BINARY)

        with pytest.raises(errors.GatewayError, match="Unexpected message type received BINARY, expected TEXT"):
            transport_impl._handle_other_message(stub_response)

    @pytest.mark.parametrize(
        "code",
        [
            *range(3990, 4000),
            errors.ShardCloseCode.DECODE_ERROR,
            errors.ShardCloseCode.INVALID_SEQ,
            errors.ShardCloseCode.UNKNOWN_ERROR,
            errors.ShardCloseCode.SESSION_TIMEOUT,
            errors.ShardCloseCode.RATE_LIMITED,
        ],
    )
    def test__handle_other_message_when_message_type_is_CLOSE_and_should_reconnect(self, code, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.CLOSE, extra="some error extra", data=code)

        with pytest.raises(errors.GatewayServerClosedConnectionError) as exinfo:
            transport_impl._handle_other_message(stub_response)

        exception = exinfo.value
        assert exception.reason == "some error extra"
        assert exception.code == int(code)
        assert exception.can_reconnect is True

    @pytest.mark.parametrize("code", [*range(4010, 4020), 5000])
    def test__handle_other_message_when_message_type_is_CLOSE_and_should_not_reconnect(self, code, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.CLOSE, extra="don't reconnect", data=code)

        with pytest.raises(errors.GatewayServerClosedConnectionError) as exinfo:
            transport_impl._handle_other_message(stub_response)

        exception = exinfo.value
        assert exception.reason == "don't reconnect"
        assert exception.code == int(code)
        assert exception.can_reconnect is False

    def test__handle_other_message_when_message_type_is_CLOSING(self, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.CLOSING)

        with pytest.raises(errors.GatewayError, match="Socket has closed"):
            transport_impl._handle_other_message(stub_response)

    def test__handle_other_message_when_message_type_is_CLOSED(self, transport_impl):
        stub_response = StubResponse(type=aiohttp.WSMsgType.CLOSED)

        with pytest.raises(errors.GatewayError, match="Socket has closed"):
            transport_impl._handle_other_message(stub_response)

    def test__handle_other_message_when_message_type_is_unknown(self, transport_impl):
        stub_response = mock.AsyncMock(return_value=StubResponse(type=aiohttp.WSMsgType.ERROR))
        exception = Exception("some error")
        transport_impl._ws.exception = mock.Mock(return_value=exception)

        with pytest.raises(errors.GatewayTransportError) as exc_info:
            transport_impl._handle_other_message(stub_response)

        assert exc_info.value.__cause__ is exception

    @pytest.mark.asyncio
    async def test__receive_and_check_text(self, transport_impl):
        transport_impl._ws.receive = mock.AsyncMock(
            return_value=StubResponse(type=aiohttp.WSMsgType.TEXT, data="some text")
        )

        assert await transport_impl._receive_and_check_text() == b"some text"

        transport_impl._ws.receive.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test__receive_and_check_text_when_message_type_is_unknown(self, transport_impl):
        transport_impl._ws.receive = mock.AsyncMock(return_value=StubResponse(type=aiohttp.WSMsgType.BINARY))

        with pytest.raises(
            errors.GatewayTransportError,
            match="Gateway transport error: Unexpected message type received BINARY, expected TEXT",
        ):
            await transport_impl._receive_and_check_text()

        transport_impl._ws.receive.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test__receive_and_check_zlib_when_payload_split_across_frames(self, transport_impl):
        response1 = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"x\xda\xf2H\xcd\xc9")
        response2 = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"\xc9W(\xcf/\xcaIQ\x04\x00\x00")
        response3 = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"\x00\xff\xff")
        transport_impl._ws.receive = mock.AsyncMock(side_effect=[response1, response2, response3])

        assert await transport_impl._receive_and_check_zlib() == b"Hello world!"

        assert transport_impl._ws.receive.call_count == 3

    @pytest.mark.asyncio
    async def test__receive_and_check_zlib_when_full_payload_in_one_frame(self, transport_impl):
        response = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"x\xdaJLD\x07\x00\x00\x00\x00\xff\xff")
        transport_impl._ws.receive = mock.AsyncMock(return_value=response)

        assert await transport_impl._receive_and_check_zlib() == b"aaaaaaaaaaaaaaaaaa"

        transport_impl._ws.receive.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test__receive_and_check_zlib_when_message_type_is_unknown(self, transport_impl):
        transport_impl._ws.receive = mock.AsyncMock(return_value=StubResponse(type=aiohttp.WSMsgType.TEXT))

        with pytest.raises(
            errors.GatewayTransportError,
            match="Gateway transport error: Unexpected message type received TEXT, expected BINARY",
        ):
            await transport_impl._receive_and_check_zlib()

    @pytest.mark.asyncio
    async def test__receive_and_check_zlib_when_issue_during_reception_of_multiple_frames(self, transport_impl):
        response1 = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"x\xda\xf2H\xcd\xc9")
        response2 = StubResponse(type=aiohttp.WSMsgType.ERROR, data="Something broke!")
        response3 = StubResponse(type=aiohttp.WSMsgType.BINARY, data=b"\x00\xff\xff")
        transport_impl._ws.receive = mock.AsyncMock(side_effect=[response1, response2, response3])
        transport_impl._ws.exception = mock.Mock(return_value=None)

        with pytest.raises(
            errors.GatewayTransportError, match=r"Gateway transport error: 'Something broke!' \[extra=None, type=258\]"
        ):
            await transport_impl._receive_and_check_zlib()

    @pytest.mark.parametrize("transport_compression", [True, False])
    @pytest.mark.asyncio
    async def test_connect(self, http_settings, proxy_settings, transport_compression):
        logger = mock.Mock()
        log_filterer = mock.Mock()
        client_session = mock.Mock()
        websocket = mock.Mock()
        loads = mock.Mock()
        dumps = mock.Mock()
        exit_stack = mock.AsyncMock(enter_async_context=mock.AsyncMock(side_effect=[client_session, websocket]))

        stack = contextlib.ExitStack()
        sleep = stack.enter_context(mock.patch.object(asyncio, "sleep"))
        create_tcp_connector = stack.enter_context(mock.patch.object(net, "create_tcp_connector"))
        create_client_session = stack.enter_context(mock.patch.object(net, "create_client_session"))
        stack.enter_context(mock.patch.object(contextlib, "AsyncExitStack", return_value=exit_stack))

        with stack:
            ws = await shard._GatewayTransport.connect(
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                logger=logger,
                url="testing.com",
                log_filterer=log_filterer,
                loads=loads,
                dumps=dumps,
                transport_compression=transport_compression,
            )

        assert isinstance(ws, shard._GatewayTransport)
        assert ws._ws is websocket
        assert ws._exit_stack is exit_stack
        assert ws._logger is logger
        assert ws._log_filterer is log_filterer
        assert ws._loads is loads
        assert ws._dumps is dumps

        if transport_compression:
            assert ws._receive_and_check == ws._receive_and_check_zlib
        else:
            assert ws._receive_and_check == ws._receive_and_check_text

        assert exit_stack.enter_async_context.call_count == 2
        exit_stack.enter_async_context.assert_has_calls(
            [mock.call(create_client_session.return_value), mock.call(client_session.ws_connect.return_value)]
        )

        create_tcp_connector.assert_called_once_with(http_settings=http_settings, dns_cache=False)
        create_client_session.assert_called_once_with(
            connector=create_tcp_connector.return_value,
            connector_owner=True,
            raise_for_status=True,
            http_settings=http_settings,
            trust_env=proxy_settings.trust_env,
        )
        client_session.ws_connect.assert_called_once_with(
            max_msg_size=0,
            proxy=proxy_settings.url,
            proxy_headers=proxy_settings.headers,
            url="testing.com",
            autoclose=False,
        )
        exit_stack.aclose.assert_not_called()
        sleep.assert_not_called()

    @pytest.mark.asyncio
    async def test_connect_when_error_while_connecting(self, http_settings, proxy_settings):
        logger = mock.Mock()
        log_filterer = mock.Mock()
        client_session = mock.Mock()
        websocket = mock.Mock()
        exit_stack = mock.AsyncMock(enter_async_context=mock.AsyncMock(side_effect=[client_session, websocket]))

        stack = contextlib.ExitStack()
        sleep = stack.enter_context(mock.patch.object(asyncio, "sleep"))
        stack.enter_context(mock.patch.object(net, "create_tcp_connector", side_effect=RuntimeError))
        stack.enter_context(mock.patch.object(contextlib, "AsyncExitStack", return_value=exit_stack))
        stack.enter_context(pytest.raises(RuntimeError))

        with stack:
            await shard._GatewayTransport.connect(
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                logger=logger,
                url="https://some.url",
                log_filterer=log_filterer,
                loads=object(),
                dumps=object(),
                transport_compression=True,
            )

        exit_stack.aclose.assert_awaited_once_with()
        sleep.assert_awaited_once_with(0.25)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ("error", "reason"),
        [
            (
                aiohttp.WSServerHandshakeError(status=123, message="some error", request_info=None, history=None),
                "WSServerHandshakeError(None, None, status=123, message='some error')",
            ),
            (aiohttp.ClientOSError("some os error"), "some os error"),
            (aiohttp.ClientConnectionError("some error"), "some error"),
            (asyncio.TimeoutError("some error"), "Timeout exceeded"),
        ],
    )
    async def test_connect_when_expected_error_while_connecting(self, http_settings, proxy_settings, error, reason):
        logger = mock.Mock()
        log_filterer = mock.Mock()
        client_session = mock.Mock()
        websocket = mock.Mock()
        exit_stack = mock.AsyncMock(enter_async_context=mock.AsyncMock(side_effect=[client_session, websocket]))

        stack = contextlib.ExitStack()
        sleep = stack.enter_context(mock.patch.object(asyncio, "sleep"))
        stack.enter_context(mock.patch.object(net, "create_tcp_connector", side_effect=error))
        stack.enter_context(mock.patch.object(contextlib, "AsyncExitStack", return_value=exit_stack))
        stack.enter_context(
            pytest.raises(errors.GatewayConnectionError, match=re.escape(f"Failed to connect to server: {reason!r}"))
        )

        with stack:
            await shard._GatewayTransport.connect(
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                logger=logger,
                url="https://some.url",
                log_filterer=log_filterer,
                transport_compression=True,
                loads=object(),
                dumps=object(),
            )

        exit_stack.aclose.assert_awaited_once_with()
        sleep.assert_awaited_once_with(0.25)


@pytest.fixture
def client(http_settings, proxy_settings):
    return shard.GatewayShardImpl(
        event_manager=mock.Mock(),
        event_factory=mock.Mock(),
        url="wss://gateway.discord.gg",
        intents=intents.Intents.ALL,
        token="lol",
        http_settings=http_settings,
        proxy_settings=proxy_settings,
    )


class TestGatewayShardImpl:
    def test__init__when_unsupported_compression_format(self):
        with pytest.raises(NotImplementedError, match=r"Unsupported compression format something"):
            shard.GatewayShardImpl(
                event_manager=mock.Mock(),
                event_factory=mock.Mock(),
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                intents=intents.Intents.ALL,
                url="wss://gaytewhuy.discord.meh",
                data_format="json",
                compression="something",
                token="12345",
            )

    def test_using_etf_is_unsupported(self, http_settings, proxy_settings):
        with pytest.raises(NotImplementedError, match="Unsupported gateway data format: etf"):
            shard.GatewayShardImpl(
                event_manager=mock.Mock(),
                event_factory=mock.Mock(),
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                token=mock.Mock(),
                url="wss://erlpack-is-broken-lol.discord.meh",
                intents=intents.Intents.ALL,
                data_format="etf",
                compression="testing",
            )

    def test_heartbeat_latency_property(self, client):
        client._heartbeat_latency = 420
        assert client.heartbeat_latency == 420

    def test_id_property(self, client):
        client._shard_id = 101
        assert client.id == 101

    def test_intents_property(self, client):
        mock_intents = object()
        client._intents = mock_intents
        assert client.intents is mock_intents

    @pytest.mark.parametrize(("keep_alive_task", "expected"), [(None, False), ("some", True)])
    def test_is_alive_property(self, client, keep_alive_task, expected):
        client._keep_alive_task = keep_alive_task

        assert client.is_alive is expected

    @pytest.mark.parametrize(
        ("ws", "handshake_event", "expected"),
        [
            (None, None, False),
            (None, True, False),
            (None, False, False),
            ("something", None, False),
            ("something", True, True),
            ("something", False, False),
        ],
    )
    def test_is_connected_property(self, client, ws, handshake_event, expected):
        client._ws = ws
        client._handshake_event = (
            None if handshake_event is None else mock.Mock(is_set=mock.Mock(return_value=handshake_event))
        )

        assert client.is_connected is expected

    def test_shard_count_property(self, client):
        client._shard_count = 69
        assert client.shard_count == 69

    def test_shard__check_if_connected_when_not_alive(self, client):
        with mock.patch.object(shard.GatewayShardImpl, "is_connected", new=False):
            with pytest.raises(errors.ComponentStateConflictError):
                client._check_if_connected()

    def test_shard__check_if_connected_when_alive(self, client):
        with mock.patch.object(shard.GatewayShardImpl, "is_connected", new=True):
            client._check_if_connected()

    @pytest.mark.parametrize("idle_since", [datetime.datetime.now(), None])
    @pytest.mark.parametrize("afk", [True, False])
    @pytest.mark.parametrize(
        "status",
        [presences.Status.DO_NOT_DISTURB, presences.Status.IDLE, presences.Status.ONLINE, presences.Status.OFFLINE],
    )
    @pytest.mark.parametrize("activity", [presences.Activity(name="foo"), None])
    def test__serialize_and_store_presence_payload_when_all_args_undefined(
        self, client, idle_since, afk, status, activity
    ):
        client._activity = activity
        client._idle_since = idle_since
        client._is_afk = afk
        client._status = status

        actual_result = client._serialize_and_store_presence_payload()

        if activity is not None:
            expected_activity = {
                "name": activity.name,
                "state": activity.state,
                "type": activity.type,
                "url": activity.url,
            }
        else:
            expected_activity = None

        if status == presences.Status.OFFLINE:
            expected_status = "invisible"
        else:
            expected_status = status.value

        expected_result = {
            "game": expected_activity,
            "since": int(idle_since.timestamp() * 1_000) if idle_since is not None else None,
            "afk": afk,
            "status": expected_status,
        }

        assert expected_result == actual_result

    @pytest.mark.parametrize("idle_since", [datetime.datetime.now(), None])
    @pytest.mark.parametrize("afk", [True, False])
    @pytest.mark.parametrize(
        "status",
        [presences.Status.DO_NOT_DISTURB, presences.Status.IDLE, presences.Status.ONLINE, presences.Status.OFFLINE],
    )
    @pytest.mark.parametrize("activity", [presences.Activity(name="foo"), None])
    def test__serialize_and_store_presence_payload_sets_state(self, client, idle_since, afk, status, activity):
        client._serialize_and_store_presence_payload(idle_since=idle_since, afk=afk, status=status, activity=activity)

        assert client._activity == activity
        assert client._idle_since == idle_since
        assert client._is_afk == afk
        assert client._status == status

    def test_get_user_id(self, client):
        client._user_id = 123

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            assert client.get_user_id() == 123

        check_if_alive.assert_called_once_with()


@pytest.mark.asyncio
class TestGatewayShardImplAsync:
    async def test_close_when_no_keep_alive_task(self, client):
        client._keep_alive_task = None

        with pytest.raises(errors.ComponentStateConflictError):
            await client.close()

    async def test_close_when_closing_event_set(self, client):
        client._keep_alive_task = mock.Mock(cancel=mock.AsyncMock())
        client._non_priority_rate_limit = mock.Mock()
        client._total_rate_limit = mock.Mock()
        client._is_closing = True

        with mock.patch.object(shard.GatewayShardImpl, "join") as join:
            await client.close()

        join.assert_awaited_once_with()
        client._keep_alive_task.cancel.assert_not_called()
        client._non_priority_rate_limit.close.assert_not_called()
        client._total_rate_limit.close.assert_not_called()

    async def test_close_when_closing_event_not_set(self, client):
        cancel_async_mock = mock.Mock()

        class TaskMock:
            def __init__(self):
                self._awaited_count = 0

            def __await__(self):
                self._awaited_count += 1
                raise asyncio.CancelledError

            @property
            def cancel(self):
                return cancel_async_mock

            def assert_awaited_once(self):
                assert self._awaited_count == 1

        client._keep_alive_task = keep_alive_task = TaskMock()
        client._non_priority_rate_limit = mock.Mock()
        client._total_rate_limit = mock.Mock()

        with mock.patch.object(shard.GatewayShardImpl, "join") as join:
            await client.close()

        join.assert_not_called()
        cancel_async_mock.assert_called_once_with()
        keep_alive_task.assert_awaited_once()
        client._non_priority_rate_limit.close.assert_called_once_with()
        client._total_rate_limit.close.assert_called_once_with()

    async def test_join_when_not_alive(self, client):
        client._keep_alive_task = None

        with pytest.raises(errors.ComponentStateConflictError):
            await client.join()

    async def test_join(self, client):
        client._keep_alive_task = object()

        with mock.patch.object(asyncio, "wait_for") as wait_for:
            with mock.patch.object(asyncio, "shield", new=mock.Mock()) as shield:
                await client.join()

        shield.assert_called_once_with(client._keep_alive_task)
        wait_for.assert_awaited_once_with(shield.return_value, timeout=None)

    async def test__send_json(self, client):
        client._total_rate_limit = mock.AsyncMock()
        client._non_priority_rate_limit = mock.AsyncMock()
        client._ws = mock.AsyncMock()
        data = object()

        await client._send_json(data)

        client._non_priority_rate_limit.acquire.assert_awaited_once_with()
        client._total_rate_limit.acquire.assert_awaited_once_with()
        client._ws.send_json.assert_awaited_once_with(data)

    async def test__send_json_when_priority(self, client):
        client._total_rate_limit = mock.AsyncMock()
        client._non_priority_rate_limit = mock.AsyncMock()
        client._ws = mock.AsyncMock()
        data = object()

        await client._send_json(data, priority=True)

        client._non_priority_rate_limit.acquire.assert_not_called()
        client._total_rate_limit.acquire.assert_awaited_once_with()
        client._ws.send_json.assert_awaited_once_with(data)

    async def test_request_guild_members_when_no_query_and_no_limit_and_GUILD_MEMBERS_not_enabled(self, client):
        client._intents = intents.Intents.GUILD_INTEGRATIONS

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(errors.MissingIntentError):
                await client.request_guild_members(123, query="", limit=0)

        check_if_alive.assert_called_once_with()

    async def test_request_guild_members_when_presences_and_GUILD_PRESENCES_not_enabled(self, client):
        client._intents = intents.Intents.GUILD_INTEGRATIONS

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(errors.MissingIntentError):
                await client.request_guild_members(123, query="test", limit=1, include_presences=True)

        check_if_alive.assert_called_once_with()

    async def test_request_guild_members_when_presences_false_and_GUILD_PRESENCES_not_enabled(self, client):
        client._intents = intents.Intents.GUILD_INTEGRATIONS

        with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
            with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
                await client.request_guild_members(123, query="test", limit=1, include_presences=False)

        send_json.assert_awaited_once_with(
            {"op": 8, "d": {"guild_id": "123", "query": "test", "presences": False, "limit": 1}}
        )

        check_if_alive.assert_called_once_with()

    @pytest.mark.parametrize("kwargs", [{"query": "some query"}, {"limit": 1}])
    async def test_request_guild_members_when_specifiying_users_with_limit_or_query(self, client, kwargs):
        client._intents = intents.Intents.GUILD_INTEGRATIONS

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(ValueError, match="Cannot specify limit/query with users"):
                await client.request_guild_members(123, users=[], **kwargs)

        check_if_alive.assert_called_once_with()

    @pytest.mark.parametrize("limit", [-1, 101])
    async def test_request_guild_members_when_limit_under_0_or_over_100(self, client, limit):
        client._intents = intents.Intents.ALL

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(ValueError, match="'limit' must be between 0 and 100, both inclusive"):
                await client.request_guild_members(123, limit=limit)

        check_if_alive.assert_called_once_with()

    async def test_request_guild_members_when_users_over_100(self, client):
        client._intents = intents.Intents.ALL

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(ValueError, match="'users' is limited to 100 users"):
                await client.request_guild_members(123, users=range(101))

        check_if_alive.assert_called_once_with()

    async def test_request_guild_members_when_nonce_over_32_chars(self, client):
        client._intents = intents.Intents.ALL

        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with pytest.raises(ValueError, match="'nonce' can be no longer than 32 byte characters long."):
                await client.request_guild_members(123, nonce="x" * 33)

        check_if_alive.assert_called_once_with()

    @pytest.mark.parametrize("include_presences", [True, False])
    async def test_request_guild_members(self, client, include_presences):
        client._intents = intents.Intents.ALL

        with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
            with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
                await client.request_guild_members(123, include_presences=include_presences)

        send_json.assert_awaited_once_with(
            {"op": 8, "d": {"guild_id": "123", "query": "", "presences": include_presences, "limit": 0}}
        )
        check_if_alive.assert_called_once_with()

    @pytest.mark.parametrize("attr", ["_keep_alive_task", "_handshake_event"])
    async def test_start_when_already_running(self, client, attr):
        setattr(client, attr, object())

        with pytest.raises(errors.ComponentStateConflictError):
            await client.start()

    async def test_start_when_shard_closed_before_starting(self, client):
        client._keep_alive_task = None
        client._shard_id = 20
        handshake_event = mock.Mock(is_set=mock.Mock(return_value=False))

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(aio, "first_completed"))
        stack.enter_context(mock.patch.object(asyncio, "shield"))
        stack.enter_context(mock.patch.object(asyncio, "create_task"))
        stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_keep_alive", new=mock.Mock()))
        stack.enter_context(mock.patch.object(asyncio, "Event", return_value=handshake_event))
        stack.enter_context(pytest.raises(RuntimeError, match="shard 20 was closed before it could start successfully"))

        with stack:
            await client.start()

        assert client._keep_alive_task is None

    async def test_start(self, client):
        client._keep_alive_task = None
        client._shard_id = 20
        handshake_event = mock.Mock(is_set=mock.Mock(return_value=True))
        keep_alive_task = mock.Mock()

        stack = contextlib.ExitStack()
        first_completed = stack.enter_context(mock.patch.object(aio, "first_completed"))
        shield = stack.enter_context(mock.patch.object(asyncio, "shield"))
        create_task = stack.enter_context(mock.patch.object(asyncio, "create_task", return_value=keep_alive_task))
        keep_alive = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_keep_alive", new=mock.Mock()))
        stack.enter_context(mock.patch.object(asyncio, "Event", return_value=handshake_event))

        with stack:
            await client.start()

        assert client._keep_alive_task is keep_alive_task

        create_task.assert_called_once_with(keep_alive.return_value, name="keep alive (shard 20)")
        shield.assert_called_once_with(create_task.return_value)
        first_completed.assert_awaited_once_with(handshake_event.wait.return_value, shield.return_value)

    async def test_update_presence(self, client):
        with mock.patch.object(shard.GatewayShardImpl, "_serialize_and_store_presence_payload") as presence:
            with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
                with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
                    await client.update_presence(
                        idle_since=datetime.datetime.now(), afk=True, status=presences.Status.IDLE, activity=None
                    )

        send_json.assert_awaited_once_with({"op": 3, "d": presence.return_value})
        check_if_alive.assert_called_once_with()

    async def test_update_voice_state(self, client):
        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
                await client.update_voice_state(123456, 6969420, self_mute=False, self_deaf=True)

        send_json.assert_awaited_once_with(
            {"op": 4, "d": {"guild_id": "123456", "channel_id": "6969420", "self_mute": False, "self_deaf": True}}
        )
        check_if_alive.assert_called_once_with()

    async def test_update_voice_state_without_optionals(self, client):
        with mock.patch.object(shard.GatewayShardImpl, "_check_if_connected") as check_if_alive:
            with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
                await client.update_voice_state(123456, 6969420)

        send_json.assert_awaited_once_with({"op": 4, "d": {"guild_id": "123456", "channel_id": "6969420"}})
        check_if_alive.assert_called_once_with()

    @hikari_test_helpers.timeout()
    async def test__heartbeat(self, client):
        client._last_heartbeat_sent = 5
        client._logger = mock.Mock()

        class ExitException(Exception): ...

        stack = contextlib.ExitStack()
        sleep = stack.enter_context(mock.patch.object(asyncio, "sleep", side_effect=[None, ExitException]))
        stack.enter_context(mock.patch.object(time, "time", return_value=10))
        send_heartbeat = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_send_heartbeat"))
        stack.enter_context(pytest.raises(ExitException))

        with stack:
            await client._heartbeat(20)

        assert send_heartbeat.await_count == 2
        send_heartbeat.assert_has_awaits([mock.call(), mock.call()])
        assert sleep.await_count == 2
        sleep.assert_has_awaits([mock.call(20), mock.call(20)])

    @hikari_test_helpers.timeout()
    async def test__heartbeat_when_zombie(self, client):
        client._last_heartbeat_sent = 10
        client._logger = mock.Mock()

        with mock.patch.object(time, "time", return_value=5):
            with mock.patch.object(asyncio, "sleep") as sleep:
                await client._heartbeat(20)

        sleep.assert_not_called()

    async def test__connect_when_ws(self, client):
        client._ws = object()

        with pytest.raises(errors.ComponentStateConflictError):
            await client._connect()

    async def test__connect_when_not_reconnecting(self, client, http_settings, proxy_settings):
        ws = mock.AsyncMock()
        ws.receive_json.return_value = {"op": 10, "d": {"heartbeat_interval": 10}}
        client._transport_compression = False
        client._shard_id = 20
        client._shard_count = 100
        client._gateway_url = "wss://somewhere.com?somewhere=true"
        client._resume_gateway_url = None
        client._token = "sometoken"
        client._logger = mock.Mock()
        client._handshake_event = mock.Mock()
        client._seq = None
        client._large_threshold = "your mom"
        client._intents = 9

        heartbeat_task = object()
        poll_events_task = object()
        shielded_heartbeat_task = object()
        shielded_poll_events_task = object()

        stack = contextlib.ExitStack()
        create_task = stack.enter_context(
            mock.patch.object(asyncio, "create_task", side_effect=[heartbeat_task, poll_events_task])
        )
        shield = stack.enter_context(
            mock.patch.object(asyncio, "shield", side_effect=[shielded_heartbeat_task, shielded_poll_events_task])
        )
        first_completed = stack.enter_context(mock.patch.object(aio, "first_completed"))
        log_filterer = stack.enter_context(mock.patch.object(shard, "_log_filterer"))
        serialize_and_store_presence_payload = stack.enter_context(
            mock.patch.object(shard.GatewayShardImpl, "_serialize_and_store_presence_payload")
        )
        send_json = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_send_json"))
        heartbeat = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_heartbeat", new=mock.Mock()))
        poll_events = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_poll_events", new=mock.Mock()))
        gateway_transport_connect = stack.enter_context(
            mock.patch.object(shard._GatewayTransport, "connect", return_value=ws)
        )
        stack.enter_context(mock.patch.object(urls, "VERSION", new=400))
        stack.enter_context(mock.patch.object(platform, "system", return_value="Potato OS"))
        stack.enter_context(mock.patch.object(platform, "architecture", return_value=["ARM64"]))
        stack.enter_context(mock.patch.object(aiohttp, "__version__", new="4.0"))
        stack.enter_context(mock.patch.object(_about, "__version__", new="1.0.0"))

        with stack:
            assert await client._connect() == (heartbeat_task, poll_events_task)

        log_filterer.assert_called_once_with(b"sometoken")
        gateway_transport_connect.assert_called_once_with(
            http_settings=http_settings,
            log_filterer=log_filterer.return_value,
            logger=client._logger,
            proxy_settings=proxy_settings,
            transport_compression=False,
            loads=client._loads,
            dumps=client._dumps,
            url="wss://somewhere.com?somewhere=true&v=400&encoding=json",
        )

        assert create_task.call_count == 2
        create_task.assert_has_calls(
            [
                mock.call(heartbeat.return_value, name="heartbeat (shard 20)"),
                mock.call(poll_events.return_value, name="poll events (shard 20)"),
            ]
        )
        heartbeat.assert_called_once_with(0.01)

        ws.receive_json.assert_awaited_once_with()
        send_json.assert_called_once_with(
            {
                "op": 2,
                "d": {
                    "token": "sometoken",
                    "compress": False,
                    "large_threshold": "your mom",
                    "properties": {
                        "os": "Potato OS ARM64",
                        "browser": "hikari (1.0.0, aiohttp 4.0)",
                        "device": "hikari 1.0.0",
                    },
                    "shard": [20, 100],
                    "intents": 9,
                    "presence": serialize_and_store_presence_payload.return_value,
                },
            }
        )

        assert shield.call_count == 2
        shield.assert_has_calls([mock.call(heartbeat_task), mock.call(poll_events_task)])
        first_completed.assert_called_once_with(
            client._handshake_event.wait.return_value, shielded_heartbeat_task, shielded_poll_events_task
        )

    async def test__connect_when_reconnecting(self, client, http_settings, proxy_settings):
        ws = mock.AsyncMock()
        ws.receive_json.return_value = {"op": 10, "d": {"heartbeat_interval": 10}}
        client._transport_compression = True
        client._shard_id = 20
        client._gateway_url = "wss://somewhere.com?somewhere=false"
        client._resume_gateway_url = "wss://notsomewhere.com?somewhere=true"
        client._token = "sometoken"
        client._logger = mock.Mock()
        client._handshake_event = mock.Mock()
        client._seq = 1234
        client._session_id = "some session id"

        heartbeat_task = object()
        poll_events_task = object()
        shielded_heartbeat_task = object()
        shielded_poll_events_task = object()

        stack = contextlib.ExitStack()
        create_task = stack.enter_context(
            mock.patch.object(asyncio, "create_task", side_effect=[heartbeat_task, poll_events_task])
        )
        shield = stack.enter_context(
            mock.patch.object(asyncio, "shield", side_effect=[shielded_heartbeat_task, shielded_poll_events_task])
        )
        first_completed = stack.enter_context(mock.patch.object(aio, "first_completed"))
        log_filterer = stack.enter_context(mock.patch.object(shard, "_log_filterer"))
        send_json = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_send_json"))
        heartbeat = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_heartbeat", new=mock.Mock()))
        poll_events = stack.enter_context(mock.patch.object(shard.GatewayShardImpl, "_poll_events", new=mock.Mock()))
        gateway_transport_connect = stack.enter_context(
            mock.patch.object(shard._GatewayTransport, "connect", return_value=ws)
        )
        stack.enter_context(mock.patch.object(urls, "VERSION", new=400))

        with stack:
            assert await client._connect() == (heartbeat_task, poll_events_task)

        log_filterer.assert_called_once_with(b"sometoken")
        gateway_transport_connect.assert_called_once_with(
            http_settings=http_settings,
            log_filterer=log_filterer.return_value,
            logger=client._logger,
            proxy_settings=proxy_settings,
            loads=client._loads,
            dumps=client._dumps,
            transport_compression=True,
            url="wss://notsomewhere.com?somewhere=true&v=400&encoding=json&compress=zlib-stream",
        )

        assert create_task.call_count == 2
        create_task.assert_has_calls(
            [
                mock.call(heartbeat.return_value, name="heartbeat (shard 20)"),
                mock.call(poll_events.return_value, name="poll events (shard 20)"),
            ]
        )
        heartbeat.assert_called_once_with(0.01)

        ws.receive_json.assert_awaited_once_with()
        send_json.assert_called_once_with(
            {"op": 6, "d": {"token": "sometoken", "seq": 1234, "session_id": "some session id"}}
        )

        assert shield.call_count == 2
        shield.assert_has_calls([mock.call(heartbeat_task), mock.call(poll_events_task)])
        first_completed.assert_called_once_with(
            client._handshake_event.wait.return_value, shielded_heartbeat_task, shielded_poll_events_task
        )

    async def test__connect_when_op_received_is_not_HELLO(self, client):
        ws = mock.AsyncMock()
        ws.receive_json.return_value = {"op": 0, "d": {"not": "hello"}}
        client._gateway_url = "somewhere.com"
        client._logger = mock.Mock()
        client._handshake_event = object()

        stack = contextlib.ExitStack()
        stack.enter_context(pytest.raises(errors.GatewayError))
        gateway_transport_connect = stack.enter_context(
            mock.patch.object(shard._GatewayTransport, "connect", return_value=ws)
        )

        with stack:
            assert await client._connect()

        gateway_transport_connect.return_value.send_close.assert_awaited_once_with(
            code=1002, message=b"Expected HELLO op"
        )

    @pytest.mark.skip("TODO")
    async def test__keep_alive(self, client): ...

    async def test__send_heartbeat(self, client):
        client._last_heartbeat_sent = 0
        client._seq = 10

        with mock.patch.object(shard.GatewayShardImpl, "_send_json") as send_json:
            with mock.patch.object(time, "time", return_value=200):
                await client._send_heartbeat()

        send_json.assert_awaited_once_with({"op": 1, "d": 10}, priority=True)
        assert client._last_heartbeat_sent == 200

    async def test__poll_events_on_dispatch(self, client):
        payload = {"op": 0, "t": "SOMETHING", "d": {"some": "test"}, "s": 101}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._seq = 1000
        client._event_manager.consume_raw_event = mock.Mock(side_effect=[LookupError])
        client._handshake_event = mock.Mock()

        with pytest.raises(RuntimeError):
            await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        assert client._seq == 101
        client._event_manager.consume_raw_event.assert_called_once_with("SOMETHING", client, {"some": "test"})
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_dispatch_when_READY(self, client):
        data = {
            "v": 10,
            "session_id": 100001,
            "resume_gateway_url": "testing_endpoint",
            "user": {"id": 123, "username": "davfsa", "discriminator": "7026"},
            "guilds": ["1"] * 100,
        }

        payload = {"op": 0, "t": "READY", "d": data, "s": 101}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._seq = None
        client._session_id = None
        client._resume_gateway_url = None
        client._user_id = None
        client._event_manager.consume_raw_event = mock.Mock(side_effect=[LookupError])
        client._handshake_event = mock.Mock()

        with pytest.raises(RuntimeError):
            await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        assert client._seq == 101
        assert client._resume_gateway_url == "testing_endpoint"
        assert client._session_id == 100001
        assert client._user_id == 123
        client._event_manager.consume_raw_event.assert_called_once_with("READY", client, data)
        client._handshake_event.set.assert_called_once_with()

    async def test__poll_events_on_dispatch_when_RESUMED(self, client):
        payload = {"op": 0, "t": "RESUMED", "d": {"some": "test"}, "s": 101}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._seq = 1000
        client._event_manager.consume_raw_event = mock.Mock(side_effect=[LookupError])
        client._handshake_event = mock.Mock()

        with pytest.raises(RuntimeError):
            await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        assert client._seq == 101
        client._event_manager.consume_raw_event.assert_called_once_with("RESUMED", client, {"some": "test"})
        client._handshake_event.set.assert_called_once_with()

    async def test__poll_events_on_heartbeat_ack(self, client):
        payload = {"op": 11}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._heartbeat_latency = 0
        client._last_heartbeat_ack_received = 0
        client._last_heartbeat_sent = 1.5
        client._handshake_event = mock.Mock()

        with mock.patch.object(time, "time", return_value=3):
            with pytest.raises(RuntimeError):
                await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        assert client._last_heartbeat_ack_received == 3
        assert client._heartbeat_latency == 1.5
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_heartbeat(self, client):
        payload = {"op": 1}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._handshake_event = mock.Mock()

        with mock.patch.object(shard.GatewayShardImpl, "_send_heartbeat") as send_heartbeat:
            with pytest.raises(RuntimeError):
                await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        send_heartbeat.assert_awaited_once_with()
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_reconnect(self, client):
        payload = {"op": 7}

        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._handshake_event = mock.Mock()

        await client._poll_events()

        assert client._ws.receive_json.await_count == 1
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_invalid_session_when_can_resume(self, client):
        payload = {"op": 9, "d": True}

        client._seq = 123
        client._session_id = 456
        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._handshake_event = mock.Mock()

        await client._poll_events()

        assert client._ws.receive_json.await_count == 1
        assert client._seq == 123
        assert client._session_id == 456
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_invalid_session_when_cant_resume(self, client):
        payload = {"op": 9, "d": False}

        client._seq = 123
        client._session_id = 456
        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._handshake_event = mock.Mock()

        await client._poll_events()

        assert client._ws.receive_json.await_count == 1
        assert client._seq is None
        assert client._session_id is None
        client._handshake_event.set.assert_not_called()

    async def test__poll_events_on_unknown_op(self, client):
        payload = {"op": 69, "d": "DATA"}

        client._logger = mock.Mock()
        client._ws = mock.Mock(receive_json=mock.AsyncMock(side_effect=[payload, RuntimeError]))
        client._handshake_event = mock.Mock()

        with pytest.raises(RuntimeError):
            await client._poll_events()

        assert client._ws.receive_json.await_count == 2
        client._logger.log.assert_called_once_with(ux.TRACE, "unknown opcode %s received, it will be ignored...", 69)
        client._handshake_event.set.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Events relating to specific shards events."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "MemberChunkEvent",
    "ShardConnectedEvent",
    "ShardDisconnectedEvent",
    "ShardEvent",
    "ShardPayloadEvent",
    "ShardReadyEvent",
    "ShardResumedEvent",
    "ShardStateEvent",
)

import abc
import typing

import attrs

from hikari.events import base_events
from hikari.internal import attrs_extensions
from hikari.internal import collections
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    from hikari import applications
    from hikari import guilds
    from hikari import presences as presences_
    from hikari import snowflakes
    from hikari import traits
    from hikari import users
    from hikari.api import shard as gateway_shard


class ShardEvent(base_events.Event, abc.ABC):
    """Base class for any event that was shard-specific."""

    __slots__: typing.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def shard(self) -> gateway_shard.GatewayShard:
        """Shard that received this event."""


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ShardPayloadEvent(ShardEvent):
    """Event fired for most shard events with their raw payload.

    !!! note
        This will only be dispatched for real dispatch events received from
        Discord and not artificial events like the [`hikari.events.shard_events.ShardStateEvent`][] events.
    """

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.

    name: str = attrs.field()
    """Name of the received event."""

    payload: typing.Mapping[str, typing.Any] = attrs.field()
    """The raw payload for this event."""


class ShardStateEvent(ShardEvent, abc.ABC):
    """Base class for any event concerning the state/connectivity of a shard.

    This currently wraps connection/disconnection/ready/resumed events only.
    """

    __slots__: typing.Sequence[str] = ()


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ShardConnectedEvent(ShardStateEvent):
    """Event fired when a shard successfully connects."""

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ShardDisconnectedEvent(ShardStateEvent):
    """Event fired when a shard disconnects."""

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ShardReadyEvent(ShardStateEvent):
    """Event fired when a shard declares it is ready."""

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.

    actual_gateway_version: int = attrs.field(repr=True)
    """Actual gateway version being used."""

    resume_gateway_url: str = attrs.field(repr=False)
    """The URL to use to when reconnecting to the gateway."""

    session_id: str = attrs.field(repr=True)
    """ID for this session."""

    my_user: users.OwnUser = attrs.field(repr=True)
    """User for the current bot account this connection is authenticated with."""

    unavailable_guilds: typing.Sequence[snowflakes.Snowflake] = attrs.field(repr=False)
    """Sequence of the IDs for all guilds this bot is currently in.

    All guilds will start off "unavailable" and should become available after
    a few seconds of connecting one-by-one.
    """

    application_id: snowflakes.Snowflake = attrs.field(repr=True)
    """ID of the application this ready event is for."""

    application_flags: applications.ApplicationFlags = attrs.field(repr=True)
    """Flags of the application this ready event is for."""

    @property
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        # <<inherited docstring from Event>>.
        return self.my_user.app


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class ShardResumedEvent(ShardStateEvent):
    """Event fired when a shard resumes an existing session."""

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.


@attrs_extensions.with_copy
@attrs.define(kw_only=True, weakref_slot=False)
class MemberChunkEvent(ShardEvent, typing.Sequence["guilds.Member"]):
    """Event fired when a member chunk payload is received on a gateway shard."""

    app: traits.RESTAware = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attrs.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<docstring inherited from ShardEvent>>.

    guild_id: snowflakes.Snowflake = attrs.field(repr=True)
    # <<docstring inherited from ShardEvent>>.

    members: typing.Mapping[snowflakes.Snowflake, guilds.Member] = attrs.field(repr=False)
    """Mapping of user IDs to the objects of the members in this chunk."""

    chunk_index: int = attrs.field(repr=True)
    """Zero-indexed position of this within the queued up chunks for this request."""

    chunk_count: int = attrs.field(repr=True)
    """Total number of expected chunks for the request this is associated with."""

    not_found: typing.Sequence[snowflakes.Snowflake] = attrs.field(repr=True)
    """Sequence of the snowflakes that were not found while making this request.

    This is only applicable when user IDs are specified while making the
    member request the chunk is associated with.
    """

    presences: typing.Mapping[snowflakes.Snowflake, presences_.MemberPresence] = attrs.field(repr=False)
    """Mapping of user IDs to found member presence objects.

    This will be empty if no presences are found or
    [`include_presences`][hikari.api.shard.GatewayShard.request_guild_members] is
    not passed as [`True`][] while requesting the member chunks.
    """

    nonce: str | None = attrs.field(repr=True)
    """String nonce used to identify the request member chunks are associated with.

    This is the nonce value passed while requesting member chunks or [`None`][]
    if there was no nonce passed.
    """

    @typing.overload
    def __getitem__(self, index_or_slice: int, /) -> guilds.Member: ...

    @typing.overload
    def __getitem__(self, index_or_slice: slice, /) -> typing.Sequence[guilds.Member]: ...

    @typing_extensions.override
    def __getitem__(self, index_or_slice: int | slice, /) -> guilds.Member | typing.Sequence[guilds.Member]:
        return collections.get_index_or_slice(self.members, index_or_slice)

    @typing_extensions.override
    def __iter__(self) -> typing.Iterator[guilds.Member]:
        return iter(self.members.values())

    @typing_extensions.override
    def __len__(self) -> int:
        return len(self.members)



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import contextlib
import signal

import mock
import pytest

from hikari import errors
from hikari.internal import signals


def test__raise_interrupt():
    with mock.patch.object(signal, "strsignal"):  # signals are not always implemented
        with pytest.raises(errors.HikariInterrupt):
            signals._raise_interrupt(1)


@pytest.mark.parametrize("trace", [True, False])
def test__interrupt_handler(trace):
    loop = mock.Mock()

    with mock.patch.object(signals, "_LOGGER", new=mock.Mock(isEnabledFor=mock.Mock(return_value=trace))):
        handler = signals._interrupt_handler(loop)

        handler(1, None)

    loop.call_soon_threadsafe.assert_called_once_with(signals._raise_interrupt, 1)


class TestHandleInterrupt:
    def test_behaviour(self):
        loop = object()

        stack = contextlib.ExitStack()
        register_signal_handler = stack.enter_context(mock.patch.object(signal, "signal"))
        interrupt_handler = stack.enter_context(mock.patch.object(signals, "_interrupt_handler"))
        stack.enter_context(mock.patch.object(signal, "SIGINT", new=2, create=True))
        stack.enter_context(mock.patch.object(signal, "SIGTERM", new=15, create=True))
        stack.enter_context(mock.patch.object(signals, "_INTERRUPT_SIGNALS", ("SIGINT", "SIGTERM", "UNIMPLEMENTED")))

        with stack:
            with signals.handle_interrupts(loop, propagate_interrupts=True, enabled=True):
                interrupt_handler.assert_called_once_with(loop)

                assert register_signal_handler.call_count == 2
                register_signal_handler.assert_has_calls(
                    [mock.call(2, interrupt_handler.return_value), mock.call(15, interrupt_handler.return_value)]
                )

                register_signal_handler.reset_mock()

        assert register_signal_handler.call_count == 2
        register_signal_handler.assert_has_calls(
            [mock.call(2, signal.default_int_handler), mock.call(15, signal.SIG_DFL)]
        )

    def test_when_disabled(self):
        with mock.patch.object(signal, "signal") as register_signal_handler:
            with signals.handle_interrupts(mock.Mock(), enabled=False, propagate_interrupts=True):
                register_signal_handler.assert_not_called()

        register_signal_handler.assert_not_called()

    def test_when_propagate_interrupt(self):
        with mock.patch.object(signal, "signal"):
            with pytest.raises(errors.HikariInterrupt):
                with signals.handle_interrupts(mock.Mock(), enabled=True, propagate_interrupts=True):
                    raise errors.HikariInterrupt(1, "t")

    def test_when_not_propagate_interrupt(self):
        with mock.patch.object(signal, "signal"):
            with signals.handle_interrupts(mock.Mock(), enabled=True, propagate_interrupts=False):
                raise errors.HikariInterrupt(1, "t")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Generate the code reference pages.

Based on https://mkdocstrings.github.io/recipes/#generate-pages-on-the-fly
"""

from __future__ import annotations

import pathlib

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

for path in sorted(pathlib.Path("hikari").rglob("*.py")):
    module_path = path.relative_to(".").with_suffix("")
    doc_path = path.relative_to(".").with_suffix(".md")
    full_doc_path = pathlib.Path("reference", doc_path)

    parts = tuple(module_path.parts)

    index = False

    # Ignore the internals module
    if "internal" in parts:
        continue
    elif parts[-1] == "__init__":
        index = True
        parts = parts[:-1]
        # Make the __init__.py the index page of the module
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")

    # Ignore dunder and private modules
    elif parts[-1].startswith("_"):
        continue

    nav[parts] = doc_path.as_posix()

    # We could probably use some sort of template file for this if needed
    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        full_name = ".".join(parts)
        fd.write(
            "---\n"
            f"title: `{full_name}`\n"
            f"description: {full_name} - API reference\n"
            "---\n"
            f"# `{full_name}`\n"
            f"::: {full_name}\n"
        )

        # As of this commit b327b908 in griffe the idea of "exported" members has changed
        # when it comes to `__init__` files, but we can work around it by explicitly
        # removing all members from the init renders, leaving only the docstrings
        # see: https://github.com/mkdocstrings/griffe/commit/b327b908d9546c8eb8f4ce5d3a216309937a6552
        # see: https://github.com/mkdocstrings/python/issues/39
        if index:
            fd.write("    options:\n")
            fd.write("      members: false\n")

    mkdocs_gen_files.set_edit_path(full_doc_path, pathlib.Path("..", path))

with mkdocs_gen_files.open("reference/summary.md", "w") as nav_file:
    for item in nav.items():
        path = pathlib.Path(item.filename).with_suffix("")

        if path.name == "index":
            path = path.parent

        full_name = ".".join(path.parts)
        nav_file.write("    " * item.level + f"* [{full_name}]({item.filename})\n")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import inspect
import sys
import typing
import warnings

import pytest


#####################
# Enable loop debug #
#####################
class TestingPolicy(asyncio.DefaultEventLoopPolicy):
    def set_event_loop(self, loop: typing.Optional[asyncio.AbstractEventLoop]) -> None:
        # Close any old event loops to prevent them from raising warnings
        if self._local._loop:
            self._local._loop.close()

        if loop is not None:
            loop.set_debug(True)

        super().set_event_loop(loop)


asyncio.set_event_loop_policy(TestingPolicy())
sys.set_coroutine_origin_tracking_depth(100)

################################################################################
# Force ids in parametrize to be stringified by default for better readability #
################################################################################
_pytest_parametrize = pytest.mark.parametrize


def parametrize(*args, **kwargs):
    kwargs.setdefault("ids", repr)
    return _pytest_parametrize(*args, **kwargs)


pytest.mark.parametrize = parametrize

#################################################
# Filter out deprecation warnings emitted by us #
#################################################

_warn = warnings.warn


def warn(*args, **kwargs):
    frame = inspect.currentframe().f_back

    if frame.f_globals.get("__name__") == "hikari.internal.deprecation" and frame.f_code.co_name == "warn_deprecated":
        # Ignore this specific DeprecationWarning
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            _warn(*args, **kwargs)

    else:
        _warn(*args, **kwargs)


warnings.warn = warn



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import sys
import warnings

import mock
import pytest

from hikari import applications
from hikari import errors
from hikari import presences
from hikari import snowflakes
from hikari import undefined
from hikari.impl import cache as cache_impl
from hikari.impl import config
from hikari.impl import entity_factory as entity_factory_impl
from hikari.impl import event_factory as event_factory_impl
from hikari.impl import event_manager as event_manager_impl
from hikari.impl import gateway_bot as bot_impl
from hikari.impl import rest as rest_impl
from hikari.impl import shard as shard_impl
from hikari.impl import voice as voice_impl
from hikari.internal import aio
from hikari.internal import signals
from hikari.internal import ux
from tests.hikari import hikari_test_helpers


@pytest.mark.parametrize("activity", [undefined.UNDEFINED, None])
def test_validate_activity_when_no_activity(activity):
    with mock.patch.object(warnings, "warn") as warn:
        bot_impl._validate_activity(activity)

    warn.assert_not_called()


def test_validate_activity_when_type_is_streaming_but_no_url():
    activity = presences.Activity(name="test", url=None, type=presences.ActivityType.STREAMING)

    with mock.patch.object(warnings, "warn") as warn:
        bot_impl._validate_activity(activity)

    warn.assert_called_once_with(
        "The STREAMING activity type requires a 'url' parameter pointing to a valid Twitch or YouTube video "
        "URL to be specified on the activity for the presence update to have any effect.",
        category=errors.HikariWarning,
        stacklevel=3,
    )


def test_validate_activity_when_no_warning():
    activity = presences.Activity(name="test", state="Hello!", type=presences.ActivityType.CUSTOM)

    with mock.patch.object(warnings, "warn") as warn:
        bot_impl._validate_activity(activity)

    warn.assert_not_called()


class TestGatewayBot:
    @pytest.fixture
    def cache(self):
        return mock.Mock()

    @pytest.fixture
    def entity_factory(self):
        return mock.Mock()

    @pytest.fixture
    def event_factory(self):
        return mock.Mock()

    @pytest.fixture
    def event_manager(self):
        return mock.Mock()

    @pytest.fixture
    def rest(self):
        return mock.Mock()

    @pytest.fixture
    def voice(self):
        return mock.Mock()

    @pytest.fixture
    def executor(self):
        return mock.Mock()

    @pytest.fixture
    def intents(self):
        return mock.Mock()

    @pytest.fixture
    def proxy_settings(self):
        return mock.Mock()

    @pytest.fixture
    def http_settings(self):
        return mock.Mock()

    @pytest.fixture
    def bot(
        self,
        cache,
        entity_factory,
        event_factory,
        event_manager,
        rest,
        voice,
        executor,
        intents,
        proxy_settings,
        http_settings,
    ):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(cache_impl, "CacheImpl", return_value=cache))
        stack.enter_context(mock.patch.object(entity_factory_impl, "EntityFactoryImpl", return_value=entity_factory))
        stack.enter_context(mock.patch.object(event_factory_impl, "EventFactoryImpl", return_value=event_factory))
        stack.enter_context(mock.patch.object(event_manager_impl, "EventManagerImpl", return_value=event_manager))
        stack.enter_context(mock.patch.object(voice_impl, "VoiceComponentImpl", return_value=voice))
        stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl", return_value=rest))
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "print_banner"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))

        with stack:
            return bot_impl.GatewayBot(
                "token",
                executor=executor,
                http_settings=http_settings,
                proxy_settings=proxy_settings,
                intents=intents,
                max_retries=0,
            )

    def test_init(self):
        stack = contextlib.ExitStack()
        cache = stack.enter_context(mock.patch.object(cache_impl, "CacheImpl"))
        entity_factory = stack.enter_context(mock.patch.object(entity_factory_impl, "EntityFactoryImpl"))
        event_factory = stack.enter_context(mock.patch.object(event_factory_impl, "EventFactoryImpl"))
        event_manager = stack.enter_context(mock.patch.object(event_manager_impl, "EventManagerImpl"))
        voice = stack.enter_context(mock.patch.object(voice_impl, "VoiceComponentImpl"))
        rest = stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl"))
        init_logging = stack.enter_context(mock.patch.object(ux, "init_logging"))
        warn_if_not_optimized = stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        print_banner = stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "print_banner"))
        executor = object()
        cache_settings = object()
        http_settings = object()
        proxy_settings = object()
        intents = object()

        with stack:
            bot = bot_impl.GatewayBot(
                "token",
                allow_color=False,
                banner="testing",
                suppress_optimization_warning=True,
                executor=executor,
                force_color=True,
                cache_settings=cache_settings,
                http_settings=http_settings,
                intents=intents,
                auto_chunk_members=False,
                logs="DEBUG",
                max_rate_limit=200,
                max_retries=0,
                proxy_settings=proxy_settings,
                rest_url="somewhere.com",
            )

        assert bot._http_settings is http_settings
        assert bot._proxy_settings is proxy_settings
        assert bot._cache is cache.return_value
        cache.assert_called_once_with(bot, cache_settings)
        assert bot._event_manager is event_manager.return_value
        event_manager.assert_called_once_with(
            entity_factory.return_value,
            event_factory.return_value,
            intents,
            auto_chunk_members=False,
            cache=cache.return_value,
        )
        assert bot._entity_factory is entity_factory.return_value
        entity_factory.assert_called_once_with(bot)
        assert bot._event_factory is event_factory.return_value
        event_factory.assert_called_once_with(bot)
        assert bot._voice is voice.return_value
        voice.assert_called_once_with(bot)
        assert bot._rest is rest.return_value
        rest.assert_called_once_with(
            cache=bot._cache,
            entity_factory=bot._entity_factory,
            executor=executor,
            http_settings=bot._http_settings,
            max_rate_limit=200,
            max_retries=0,
            proxy_settings=bot._proxy_settings,
            dumps=bot._dumps,
            loads=bot._loads,
            rest_url="somewhere.com",
            token="token",
            token_type=applications.TokenType.BOT,
        )

        init_logging.assert_called_once_with("DEBUG", allow_color=False, force_color=True)
        print_banner.assert_called_once_with("testing", allow_color=False, force_color=True)
        warn_if_not_optimized.assert_called_once_with(suppress=True)

    def test_init_when_no_settings(self):
        stack = contextlib.ExitStack()
        cache = stack.enter_context(mock.patch.object(cache_impl, "CacheImpl"))
        stack.enter_context(mock.patch.object(entity_factory_impl, "EntityFactoryImpl"))
        stack.enter_context(mock.patch.object(event_factory_impl, "EventFactoryImpl"))
        stack.enter_context(mock.patch.object(event_manager_impl, "EventManagerImpl"))
        stack.enter_context(mock.patch.object(voice_impl, "VoiceComponentImpl"))
        stack.enter_context(mock.patch.object(rest_impl, "RESTClientImpl"))
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "print_banner"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))
        http_settings = stack.enter_context(mock.patch.object(config, "HTTPSettings"))
        proxy_settings = stack.enter_context(mock.patch.object(config, "ProxySettings"))
        cache_settings = stack.enter_context(mock.patch.object(config, "CacheSettings"))

        with stack:
            bot = bot_impl.GatewayBot("token", cache_settings=None, http_settings=None, proxy_settings=None)

        assert bot._http_settings is http_settings.return_value
        http_settings.assert_called_once_with()
        assert bot._proxy_settings is proxy_settings.return_value
        proxy_settings.assert_called_once_with()
        cache.assert_called_once_with(bot, cache_settings.return_value)
        cache_settings.assert_called_once_with()

    def test_init_strips_token(self):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(ux, "init_logging"))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "print_banner"))
        stack.enter_context(mock.patch.object(ux, "warn_if_not_optimized"))

        with stack:
            bot = bot_impl.GatewayBot(
                "\n\r token yeet \r\n", cache_settings=None, http_settings=None, proxy_settings=None
            )

        assert bot._token == "token yeet"

    def test_cache(self, bot, cache):
        assert bot.cache is cache

    def test_event_manager(self, bot, event_manager):
        assert bot.event_manager is event_manager

    def test_entity_factory(self, bot, entity_factory):
        assert bot.entity_factory is entity_factory

    def test_event_factory(self, bot, event_factory):
        assert bot.event_factory is event_factory

    def test_executor(self, bot, executor):
        assert bot.executor is executor

    def test_heartbeat_latencies(self, bot):
        bot._shards = {
            0: mock.Mock(id=0, heartbeat_latency=96),
            1: mock.Mock(id=1, heartbeat_latency=123),
            2: mock.Mock(id=2, heartbeat_latency=456),
        }

        assert bot.heartbeat_latencies == {0: 96, 1: 123, 2: 456}

    def test_heartbeat_latency(self, bot):
        bot._shards = {
            0: mock.Mock(heartbeat_latency=96),
            1: mock.Mock(heartbeat_latency=123),
            2: mock.Mock(heartbeat_latency=float("nan")),
        }

        assert bot.heartbeat_latency == 109.5

    def test_http_settings(self, bot, http_settings):
        assert bot.http_settings is http_settings

    def test_intents(self, bot, intents):
        assert bot.intents is intents

    def test_get_me(self, bot, cache):
        assert bot.get_me() is cache.get_me.return_value

    def test_proxy_settings(self, bot, proxy_settings):
        assert bot.proxy_settings is proxy_settings

    def test_shard_count_when_no_shards(self, bot):
        bot._shards = {}

        assert bot.shard_count == 0

    def test_shard_count(self, bot):
        bot._shards = {0: mock.Mock(shard_count=96), 1: mock.Mock(shard_count=123)}

        assert bot.shard_count == 96

    def test_voice(self, bot, voice):
        assert bot.voice is voice

    def test_rest(self, bot, rest):
        assert bot.rest is rest

    @pytest.mark.parametrize(("closed_event", "expected"), [("something", True), (None, False)])
    def test_is_alive(self, bot, closed_event, expected):
        bot._closed_event = closed_event

        assert bot.is_alive is expected

    def test_check_if_alive(self, bot):
        bot._closed_event = object()

        bot._check_if_alive()

    def test_check_if_alive_when_False(self, bot):
        bot._closed_event = None

        with pytest.raises(errors.ComponentStateConflictError):
            bot._check_if_alive()

    @pytest.mark.asyncio
    async def test_close_when_already_closed(self, bot):
        bot._closed_event = mock.Mock()

        with pytest.raises(errors.ComponentStateConflictError):
            await bot.close()

    @pytest.mark.asyncio
    async def test_close_when_already_closing(self, bot):
        bot._closed_event = mock.Mock()
        bot._closing_event = mock.Mock(is_set=mock.Mock(return_value=True))

        with mock.patch.object(bot_impl.GatewayBot, "join") as join:
            await bot.close()

        join.assert_awaited_once_with()
        bot._closed_event.set.assert_not_called()

    @pytest.mark.asyncio
    async def test_close(self, bot, event_manager, event_factory, rest, voice, cache):
        def null_call(arg):
            return arg

        class AwaitableMock:
            def __init__(self, error=None):
                self._awaited_count = 0
                self._error = error

            def __await__(self):
                if False:
                    yield  # Turns it into a generator

                self._awaited_count += 1

                if self._error:
                    raise self._error

            def __call__(self):
                return self

            def assert_awaited_once(self):
                assert self._awaited_count == 1

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(asyncio, "as_completed", side_effect=null_call))
        ensure_future = stack.enter_context(mock.patch.object(asyncio, "ensure_future", side_effect=null_call))
        get_running_loop = stack.enter_context(mock.patch.object(asyncio, "get_running_loop"))
        mock_future = mock.Mock()
        get_running_loop.return_value.create_future.return_value = mock_future

        event_manager.dispatch = mock.AsyncMock()
        rest.close = AwaitableMock()
        voice.close = AwaitableMock()
        bot._closed_event = close_event = mock.Mock()
        bot._closing_event = closing_event = mock.Mock(is_set=mock.Mock(return_value=False))
        error = RuntimeError()
        shard0 = mock.Mock(id=0, close=AwaitableMock())
        shard1 = mock.Mock(id=1, close=AwaitableMock(error))
        shard2 = mock.Mock(id=2, close=AwaitableMock())
        bot._shards = {0: shard0, 1: shard1, 2: shard2}

        with stack:
            await bot.close()

        # Events and args
        close_event.set.assert_called_once_with()
        assert bot._closed_event is None

        closing_event.set.assert_called_once_with()
        assert bot._closing_event is None

        # Closing components
        ensure_future.assert_has_calls(
            [
                mock.call(voice.close()),
                mock.call(shard0.close()),
                mock.call(shard1.close()),
                mock.call(shard2.close()),
                mock.call(rest.close()),
            ]
        )

        rest.close.assert_awaited_once()
        voice.close.assert_awaited_once()
        shard0.close.assert_awaited_once()
        shard1.close.assert_awaited_once()
        shard2.close.assert_awaited_once()

        # Error handling
        get_running_loop.assert_called_once_with()
        get_running_loop.return_value.call_exception_handler.assert_called_once_with(
            {"message": "shard 1 raised an exception during shut down", "future": shard1.close(), "exception": error}
        )

        # Clear out maps
        assert bot._shards == {}
        cache.clear.assert_called_once_with()

        event_manager.dispatch.assert_has_awaits(
            [
                mock.call(event_factory.deserialize_stopping_event.return_value, return_tasks=True),
                mock.call(event_factory.deserialize_stopped_event.return_value, return_tasks=True),
            ]
        )

    def test_dispatch(self, bot, event_manager):
        event = mock.Mock()

        assert bot.dispatch(event, return_tasks=True) is event_manager.dispatch.return_value

        event_manager.dispatch.assert_called_once_with(event, return_tasks=True)

    def test_get_listeners(self, bot, event_manager):
        event = object()

        assert bot.get_listeners(event, polymorphic=False) is event_manager.get_listeners.return_value

        event_manager.get_listeners.assert_called_once_with(event, polymorphic=False)

    @pytest.mark.asyncio
    async def test_join(self, bot, event_manager):
        bot._closed_event = mock.AsyncMock()

        await bot.join()

        bot._closed_event.wait.assert_awaited_once_with()

    @pytest.mark.asyncio
    async def test_join_when_not_running(self, bot, event_manager):
        bot._closed_event = None

        with pytest.raises(errors.ComponentStateConflictError):
            await bot.join()

    def test_listen(self, bot, event_manager):
        event = object()

        assert bot.listen(event) is event_manager.listen.return_value

        event_manager.listen.assert_called_once_with(event)

    def test_print_banner(self, bot):
        with mock.patch.object(ux, "print_banner") as print_banner:
            bot.print_banner("testing", allow_color=True, force_color=False, extra_args={"test_key": "test_value"})

        print_banner.assert_called_once_with(
            "testing", allow_color=True, force_color=False, extra_args={"test_key": "test_value"}
        )

    def test_run_when_already_running(self, bot):
        bot._closed_event = object()

        with pytest.raises(errors.ComponentStateConflictError):
            bot.run()

    def test_run_when_shard_ids_specified_without_shard_count(self, bot):
        with pytest.raises(TypeError, match=r"'shard_ids' must be passed with 'shard_count'"):
            bot.run(shard_ids={1})

    def test_run_with_asyncio_debug(self, bot):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "start", new=mock.Mock()))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "join", new=mock.Mock()))
        stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        loop = stack.enter_context(mock.patch.object(aio, "get_or_make_loop")).return_value

        with stack:
            bot.run(close_loop=False, asyncio_debug=True)

        loop.set_debug.assert_called_once_with(True)

    def test_run_with_coroutine_tracking_depth(self, bot):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "start", new=mock.Mock()))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "join", new=mock.Mock()))
        stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        stack.enter_context(mock.patch.object(aio, "get_or_make_loop"))
        coroutine_tracking_depth = stack.enter_context(
            mock.patch.object(sys, "set_coroutine_origin_tracking_depth", create=True, side_effect=AttributeError)
        )

        with stack:
            bot.run(close_loop=False, coroutine_tracking_depth=100)

        coroutine_tracking_depth.assert_called_once_with(100)

    def test_run_with_close_passed_executor(self, bot):
        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "start", new=mock.Mock()))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "join", new=mock.Mock()))
        stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        stack.enter_context(mock.patch.object(aio, "get_or_make_loop"))
        executor = mock.Mock()
        bot._executor = executor

        with stack:
            bot.run(close_loop=False, close_passed_executor=True)

        executor.shutdown.assert_called_once_with(wait=True)
        assert bot._executor is None

    def test_run_when_close_loop(self, bot):
        stack = contextlib.ExitStack()
        logger = stack.enter_context(mock.patch.object(bot_impl, "_LOGGER"))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "start", new=mock.Mock()))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "join", new=mock.Mock()))
        stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        destroy_loop = stack.enter_context(mock.patch.object(aio, "destroy_loop"))
        loop = stack.enter_context(mock.patch.object(aio, "get_or_make_loop")).return_value

        with stack:
            bot.run(close_loop=True)

        destroy_loop.assert_called_once_with(loop, logger)

    def test_run(self, bot):
        activity = object()
        afk = object()
        check_for_updates = object()
        idle_since = object()
        ignore_session_start_limit = object()
        large_threshold = object()
        shard_ids = object()
        shard_count = object()
        status = object()

        stack = contextlib.ExitStack()
        start_function = stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "start", new=mock.Mock()))
        join_function = stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "join", new=mock.Mock()))
        handle_interrupts = stack.enter_context(
            mock.patch.object(signals, "handle_interrupts", return_value=hikari_test_helpers.ContextManagerMock())
        )
        loop = stack.enter_context(mock.patch.object(aio, "get_or_make_loop")).return_value

        with stack:
            bot.run(
                activity=activity,
                afk=afk,
                asyncio_debug=False,
                check_for_updates=check_for_updates,
                close_passed_executor=False,
                close_loop=False,
                coroutine_tracking_depth=None,
                enable_signal_handlers=False,
                idle_since=idle_since,
                ignore_session_start_limit=ignore_session_start_limit,
                large_threshold=large_threshold,
                propagate_interrupts=False,
                shard_ids=shard_ids,
                shard_count=shard_count,
                status=status,
                startup_window_delay=15,
            )

        loop.run_until_complete.assert_has_calls(
            [mock.call(start_function.return_value), mock.call(join_function.return_value)]
        )
        start_function.assert_called_once_with(
            activity=activity,
            afk=afk,
            check_for_updates=check_for_updates,
            idle_since=idle_since,
            ignore_session_start_limit=ignore_session_start_limit,
            large_threshold=large_threshold,
            shard_ids=shard_ids,
            shard_count=shard_count,
            status=status,
            startup_window_delay=15,
        )
        handle_interrupts.assert_called_once_with(enabled=False, loop=loop, propagate_interrupts=False)
        handle_interrupts.return_value.assert_used_once()

    @pytest.mark.asyncio
    async def test_start_when_shard_ids_specified_without_shard_count(self, bot):
        with pytest.raises(TypeError, match=r"'shard_ids' must be passed with 'shard_count'"):
            await bot.start(shard_ids=(1,))

    @pytest.mark.asyncio
    async def test_start_when_already_running(self, bot):
        bot._closed_event = object()

        with pytest.raises(errors.ComponentStateConflictError):
            await bot.start()

    @pytest.mark.asyncio
    async def test_start(self, bot, rest, voice, event_manager, event_factory, http_settings, proxy_settings):
        class MockSessionStartLimit:
            remaining = 10
            reset_at = "now"
            max_concurrency = 1

        class MockInfo:
            url = "yourmom.eu"
            shard_count = 2
            session_start_limit = MockSessionStartLimit()

        shard1 = mock.Mock()
        shard2 = mock.Mock()
        shards_iter = iter((shard1, shard2))

        mock_start_one_shard = mock.Mock()

        def _mock_start_one_shard(*args, **kwargs):
            bot._shards[kwargs["shard_id"]] = next(shards_iter)
            return mock_start_one_shard(*args, **kwargs)

        stack = contextlib.ExitStack()
        stack.enter_context(mock.patch.object(bot_impl, "_validate_activity"))
        stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "_start_one_shard", new=_mock_start_one_shard))
        create_task = stack.enter_context(mock.patch.object(asyncio, "create_task"))
        gather = stack.enter_context(mock.patch.object(asyncio, "gather"))
        event = stack.enter_context(mock.patch.object(asyncio, "Event"))
        first_completed = stack.enter_context(
            mock.patch.object(aio, "first_completed", side_effect=[None, asyncio.TimeoutError, None])
        )
        check_for_updates = stack.enter_context(mock.patch.object(ux, "check_for_updates", new=mock.Mock()))

        event_manager.dispatch = mock.AsyncMock()
        rest.fetch_gateway_bot_info = mock.AsyncMock(return_value=MockInfo())

        with stack:
            await bot.start(
                check_for_updates=True,
                shard_ids=(2, 10),
                shard_count=20,
                activity="some activity",
                afk=True,
                idle_since="some idle since",
                status="some status",
                large_threshold=500,
                startup_window_delay=10,
            )

        check_for_updates.assert_called_once_with(http_settings, proxy_settings)
        create_task.assert_called_once_with(check_for_updates.return_value, name="check for package updates")
        rest.start.assert_called_once_with()
        voice.start.assert_called_once_with()

        assert event_manager.dispatch.call_count == 2
        event_manager.dispatch.assert_has_awaits(
            [
                mock.call(event_factory.deserialize_starting_event.return_value, return_tasks=True),
                mock.call(event_factory.deserialize_started_event.return_value, return_tasks=True),
            ]
        )

        assert gather.call_count == 2
        gather.assert_has_calls(
            [mock.call(mock_start_one_shard.return_value), mock.call(mock_start_one_shard.return_value)]
        )

        assert mock_start_one_shard.call_count == 2
        mock_start_one_shard.assert_has_calls(
            [
                mock.call(
                    bot,
                    activity="some activity",
                    afk=True,
                    idle_since="some idle since",
                    status="some status",
                    large_threshold=500,
                    shard_id=i,
                    shard_count=20,
                    url="yourmom.eu",
                )
                for i in (2, 10)
            ]
        )

        closing_event_closing_wait = event.return_value.wait.return_value
        assert first_completed.await_count == 3

        first_completed.assert_has_awaits(
            [
                # Shard 1
                mock.call(closing_event_closing_wait, gather.return_value),
                # Shard 2
                mock.call(closing_event_closing_wait, shard1.join.return_value, timeout=10),
                mock.call(closing_event_closing_wait, gather.return_value),
            ]
        )

    @pytest.mark.asyncio
    async def test_start_when_request_close_mid_startup(self, bot, rest, voice, event_manager, event_factory):
        class MockSessionStartLimit:
            remaining = 10
            reset_at = "now"
            max_concurrency = 1

        class MockInfo:
            url = "yourmom.eu"
            shard_count = 2
            session_start_limit = MockSessionStartLimit()

        # Assume that we already started one shard
        shard1 = mock.Mock()
        bot._shards = {"1": shard1}

        stack = contextlib.ExitStack()
        start_one_shard = stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "_start_one_shard"))
        first_completed = stack.enter_context(mock.patch.object(aio, "first_completed"))
        event = stack.enter_context(
            mock.patch.object(asyncio, "Event", return_value=mock.Mock(is_set=mock.Mock(return_value=True)))
        )

        event_manager.dispatch = mock.AsyncMock()
        rest.fetch_gateway_bot_info = mock.AsyncMock(return_value=MockInfo())

        with stack:
            await bot.start(shard_ids=(2, 10), shard_count=20, check_for_updates=False)

        start_one_shard.assert_not_called()

        first_completed.assert_called_once_with(
            event.return_value.wait.return_value, shard1.join.return_value, timeout=5
        )

    @pytest.mark.asyncio
    async def test_start_when_shard_closed_mid_startup(self, bot, rest, voice, event_manager, event_factory):
        class MockSessionStartLimit:
            remaining = 10
            reset_at = "now"
            max_concurrency = 1

        class MockInfo:
            url = "yourmom.eu"
            shard_count = 2
            session_start_limit = MockSessionStartLimit()

        # Assume that we already started one shard
        shard1 = mock.Mock()
        bot._shards = {"1": shard1}

        stack = contextlib.ExitStack()
        start_one_shard = stack.enter_context(mock.patch.object(bot_impl.GatewayBot, "_start_one_shard"))
        first_completed = stack.enter_context(mock.patch.object(aio, "first_completed"))
        event = stack.enter_context(
            mock.patch.object(asyncio, "Event", return_value=mock.Mock(is_set=mock.Mock(return_value=False)))
        )
        stack.enter_context(pytest.raises(RuntimeError, match="One or more shards closed while starting"))

        event_manager.dispatch = mock.AsyncMock()
        rest.fetch_gateway_bot_info = mock.AsyncMock(return_value=MockInfo())

        with stack:
            await bot.start(shard_ids=(2, 10), shard_count=20, check_for_updates=False)

        start_one_shard.assert_not_called()

        first_completed.assert_called_once_with(
            event.return_value.wait.return_value, shard1.join.return_value, timeout=5
        )

    def test_stream(self, bot):
        event_type = object()

        with mock.patch.object(bot_impl.GatewayBot, "_check_if_alive") as check_if_alive:
            bot.stream(event_type, timeout=100, limit=400)

        check_if_alive.assert_called_once_with()
        bot._event_manager.stream.assert_called_once_with(event_type, timeout=100, limit=400)

    def test_subscribe(self, bot):
        event_type = object()
        callback = object()

        bot.subscribe(event_type, callback)

        bot._event_manager.subscribe.assert_called_once_with(event_type, callback)

    def test_unsubscribe(self, bot):
        event_type = object()
        callback = object()

        bot.unsubscribe(event_type, callback)

        bot._event_manager.unsubscribe.assert_called_once_with(event_type, callback)

    @pytest.mark.asyncio
    async def test_wait_for(self, bot):
        event_type = object()
        predicate = object()
        bot._event_manager.wait_for = mock.AsyncMock()

        with mock.patch.object(bot_impl.GatewayBot, "_check_if_alive") as check_if_alive:
            await bot.wait_for(event_type, timeout=100, predicate=predicate)

        check_if_alive.assert_called_once_with()
        bot._event_manager.wait_for.assert_awaited_once_with(event_type, timeout=100, predicate=predicate)

    def test_get_shard_when_not_present(self, bot):
        shard = mock.Mock(shard_count=96)
        bot._shards = {96: shard}

        with mock.patch.object(snowflakes, "calculate_shard_id", return_value=0) as calculate_shard_id:
            with pytest.raises(
                RuntimeError, match=r"Guild 702763150025556029 isn't covered by any of the shards in this client"
            ):
                bot._get_shard(702763150025556029)

        calculate_shard_id.assert_called_once_with(96, 702763150025556029)

    def test_get_shard(self, bot):
        shard = mock.Mock(shard_count=96)
        bot._shards = {96: shard}

        with mock.patch.object(snowflakes, "calculate_shard_id", return_value=96) as calculate_shard_id:
            assert bot._get_shard(702763150025556029) is shard

        calculate_shard_id.assert_called_once_with(96, 702763150025556029)

    @pytest.mark.asyncio
    async def test_update_presence(self, bot):
        status = object()
        activity = object()
        idle_since = object()
        afk = object()

        shard0 = mock.Mock()
        shard1 = mock.Mock()
        shard2 = mock.Mock()
        bot._shards = {0: shard0, 1: shard1, 2: shard2}

        with mock.patch.object(bot_impl.GatewayBot, "_check_if_alive") as check_if_alive:
            with mock.patch.object(aio, "all_of") as all_of:
                with mock.patch.object(bot_impl, "_validate_activity") as validate_activity:
                    await bot.update_presence(status=status, activity=activity, idle_since=idle_since, afk=afk)

        check_if_alive.assert_called_once_with()
        validate_activity.assert_called_once_with(activity)
        all_of.assert_awaited_once_with(
            shard0.update_presence.return_value,
            shard1.update_presence.return_value,
            shard2.update_presence.return_value,
        )
        shard0.update_presence.assert_called_once_with(status=status, activity=activity, idle_since=idle_since, afk=afk)
        shard1.update_presence.assert_called_once_with(status=status, activity=activity, idle_since=idle_since, afk=afk)
        shard2.update_presence.assert_called_once_with(status=status, activity=activity, idle_since=idle_since, afk=afk)

    @pytest.mark.asyncio
    async def test_update_voice_state(self, bot):
        shard = mock.Mock()
        shard.update_voice_state = mock.AsyncMock()

        with mock.patch.object(bot_impl.GatewayBot, "_get_shard", return_value=shard) as get_shard:
            with mock.patch.object(bot_impl.GatewayBot, "_check_if_alive") as check_if_alive:
                await bot.update_voice_state(115590097100865541, 123, self_mute=True, self_deaf=False)

        check_if_alive.assert_called_once_with()
        get_shard.assert_called_once_with(115590097100865541)
        shard.update_voice_state.assert_awaited_once_with(
            guild=115590097100865541, channel=123, self_mute=True, self_deaf=False
        )

    @pytest.mark.asyncio
    async def test_request_guild_members(self, bot):
        shard = mock.Mock(shard_count=3)
        shard.request_guild_members = mock.AsyncMock()

        with mock.patch.object(bot_impl.GatewayBot, "_get_shard", return_value=shard) as get_shard:
            with mock.patch.object(bot_impl.GatewayBot, "_check_if_alive") as check_if_alive:
                await bot.request_guild_members(
                    115590097100865541, include_presences=True, query="indeed", limit=42, users=[123], nonce="NONCE"
                )

        check_if_alive.assert_called_once_with()
        get_shard.assert_called_once_with(115590097100865541)
        shard.request_guild_members.assert_awaited_once_with(
            guild=115590097100865541, include_presences=True, query="indeed", limit=42, users=[123], nonce="NONCE"
        )

    @pytest.mark.asyncio
    async def test_start_one_shard(self, bot):
        activity = object()
        status = object()
        bot._shards = {}
        shard_obj = mock.Mock(is_alive=True, start=mock.AsyncMock())

        with mock.patch.object(shard_impl, "GatewayShardImpl", new=mock.Mock(return_value=shard_obj)) as shard:
            await bot._start_one_shard(
                activity=activity,
                afk=True,
                idle_since=None,
                status=status,
                large_threshold=1000,
                shard_id=1,
                shard_count=3,
                url="https://some.website",
            )

        shard.assert_called_once_with(
            http_settings=bot._http_settings,
            proxy_settings=bot._proxy_settings,
            event_manager=bot._event_manager,
            event_factory=bot._event_factory,
            intents=bot._intents,
            initial_activity=activity,
            initial_is_afk=True,
            initial_idle_since=None,
            initial_status=status,
            large_threshold=1000,
            shard_id=1,
            shard_count=3,
            loads=bot._loads,
            dumps=bot._dumps,
            token=bot._token,
            url="https://some.website",
        )
        shard_obj.start.assert_awaited_once_with()
        assert bot._shards == {1: shard_obj}

    @pytest.mark.asyncio
    async def test_start_one_shard_when_not_alive(self, bot):
        activity = object()
        status = object()
        bot._shards = {}
        shard_obj = mock.Mock(is_alive=False, start=mock.AsyncMock())

        with mock.patch.object(shard_impl, "GatewayShardImpl", return_value=shard_obj):
            with pytest.raises(RuntimeError, match=r"shard 1 shut down immediately when starting"):
                await bot._start_one_shard(
                    activity=activity,
                    afk=True,
                    idle_since=None,
                    status=status,
                    large_threshold=1000,
                    shard_id=1,
                    shard_count=3,
                    url="https://some.website",
                )

        assert bot._shards == {}
        shard_obj.close.assert_not_called()

    @pytest.mark.parametrize("is_alive", [True, False])
    @pytest.mark.asyncio
    async def test_start_one_shard_when_exception(self, bot, is_alive):
        activity = object()
        status = object()
        bot._shards = {}
        shard_obj = mock.Mock(
            is_alive=is_alive, start=mock.AsyncMock(side_effect=RuntimeError("exit in tests")), close=mock.AsyncMock()
        )

        with mock.patch.object(shard_impl, "GatewayShardImpl", return_value=shard_obj):
            with pytest.raises(RuntimeError, match=r"exit in tests"):
                await bot._start_one_shard(
                    activity=activity,
                    afk=True,
                    idle_since=None,
                    status=status,
                    large_threshold=1000,
                    shard_id=1,
                    shard_count=3,
                    url="https://some.website",
                )

        assert bot._shards == {}

        if is_alive:
            shard_obj.close.assert_awaited_once_with()
        else:
            shard_obj.close.assert_not_called()



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import asyncio
import contextlib
import math
import sys
import time

import mock
import pytest

from hikari.impl import rate_limits
from tests.hikari import hikari_test_helpers


class MockFuture(mock.Mock):
    def __await__(self):
        if False:
            yield  # Turns this into a generator.
        return None


class TestBaseRateLimiter:
    def test_context_management(self):
        class MockedBaseRateLimiter(rate_limits.BaseRateLimiter):
            close = mock.Mock()
            acquire = NotImplemented

        with MockedBaseRateLimiter() as m:
            pass

        m.close.assert_called_once()


class TestBurstRateLimiter:
    @pytest.fixture
    def mock_burst_limiter(self):
        class Impl(rate_limits.BurstRateLimiter):
            async def acquire(self, *args, **kwargs) -> None:
                raise NotImplementedError

        return Impl(__name__)

    @pytest.mark.parametrize(("queue", "is_empty"), [(["foo", "bar", "baz"], False), ([], True)])
    def test_is_empty(self, queue, is_empty, mock_burst_limiter):
        mock_burst_limiter.queue = queue
        assert mock_burst_limiter.is_empty is is_empty

    @pytest.mark.asyncio
    async def test_close_removes_all_futures_from_queue(self, mock_burst_limiter):
        event_loop = asyncio.get_running_loop()
        mock_burst_limiter.throttle_task = None
        futures = [event_loop.create_future() for _ in range(10)]
        mock_burst_limiter.queue = list(futures)
        mock_burst_limiter.close()
        assert len(mock_burst_limiter.queue) == 0

    @pytest.mark.asyncio
    async def test_close_cancels_all_futures_pending_when_futures_pending(self, mock_burst_limiter):
        event_loop = asyncio.get_running_loop()
        mock_burst_limiter.throttle_task = None
        futures = [event_loop.create_future() for _ in range(10)]
        mock_burst_limiter.queue = list(futures)
        mock_burst_limiter.close()
        for i, future in enumerate(futures):
            assert future.cancelled(), f"future {i} was not cancelled"

    @pytest.mark.asyncio
    async def test_close_is_silent_when_no_futures_pending(self, mock_burst_limiter):
        mock_burst_limiter.throttle_task = None
        mock_burst_limiter.queue = []
        mock_burst_limiter.close()
        assert True, "passed successfully"

    @pytest.mark.asyncio
    async def test_close_cancels_throttle_task_if_running(self, mock_burst_limiter):
        event_loop = asyncio.get_running_loop()
        task = event_loop.create_future()
        mock_burst_limiter.throttle_task = task
        mock_burst_limiter.close()
        assert mock_burst_limiter.throttle_task is None, "task was not overwritten with None"
        assert task.cancelled(), "throttle_task is not cancelled"

    @pytest.mark.asyncio
    async def test_close_when_closed(self, mock_burst_limiter):
        # Double-running shouldn't do anything adverse.
        mock_burst_limiter.close()
        mock_burst_limiter.close()


class TestManualRateLimiter:
    @pytest.mark.asyncio
    async def test_acquire_returns_completed_future_if_throttle_task_is_None(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.ManualRateLimiter() as limiter:
            limiter.throttle_task = None
            future = MockFuture()
            event_loop.create_future = mock.Mock(return_value=future)

            await limiter.acquire()
            future.set_result.assert_called_once_with(None)

    @pytest.mark.asyncio
    async def test_acquire_returns_incomplete_future_if_throttle_task_is_not_None(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.ManualRateLimiter() as limiter:
            limiter.throttle_task = event_loop.create_future()
            future = MockFuture()
            event_loop.create_future = mock.Mock(return_value=future)

            await limiter.acquire()
            future.set_result.assert_not_called()

    @pytest.mark.asyncio
    async def test_acquire_places_future_on_queue_if_throttle_task_is_not_None(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.ManualRateLimiter() as limiter:
            limiter.throttle_task = event_loop.create_future()
            future = MockFuture()
            event_loop.create_future = mock.Mock(return_value=future)

            assert len(limiter.queue) == 0

            await limiter.acquire()

            assert len(limiter.queue) == 1
            assert future in limiter.queue
            future.set_result.assert_not_called()

    @pytest.mark.asyncio
    async def test_throttle_cancels_existing_task(self):
        with rate_limits.ManualRateLimiter() as limiter:
            limiter.throttle_task = asyncio.get_running_loop().create_future()
            old_task = limiter.throttle_task
            limiter.throttle(0)
            assert old_task.cancelled()
            assert old_task is not limiter.throttle_task

    @pytest.mark.asyncio
    async def test_throttle_schedules_throttle(self):
        with hikari_test_helpers.mock_class_namespace(rate_limits.ManualRateLimiter, slots_=False)() as limiter:
            limiter.unlock_later = mock.AsyncMock()
            limiter.throttle(0)
            await limiter.throttle_task
            limiter.unlock_later.assert_called_once_with(0)

    @pytest.mark.asyncio
    async def test_throttle_chews_queue_completing_futures(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.ManualRateLimiter() as limiter:
            futures = [event_loop.create_future() for _ in range(10)]
            limiter.queue = list(futures)
            await limiter.unlock_later(0.01)
            for i, future in enumerate(futures):
                assert future.done(), f"future {i} was not done"

    @pytest.mark.asyncio
    async def test_throttle_sleeps_before_popping_queue(self):
        event_loop = asyncio.get_running_loop()
        # GIVEN
        slept_at = float("nan")
        popped_at = []

        async def mock_sleep(_):
            nonlocal slept_at
            slept_at = time.time()

        class MockList(list):
            def pop(self, _=-1):
                popped_at.append(time.time())
                return event_loop.create_future()

        with hikari_test_helpers.mock_class_namespace(rate_limits.ManualRateLimiter, slots_=False)() as limiter:
            with mock.patch("asyncio.sleep", wraps=mock_sleep):
                limiter.queue = MockList()

                # WHEN
                await limiter.unlock_later(5)

        # THEN
        for i, pop_time in enumerate(popped_at):
            assert slept_at < pop_time, f"future {i} popped before initial sleep"

    @pytest.mark.asyncio
    async def test_throttle_clears_throttle_task(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.ManualRateLimiter() as limiter:
            limiter.throttle_task = event_loop.create_future()
            await limiter.unlock_later(0)
        assert limiter.throttle_task is None


class TestWindowedBurstRateLimiter:
    @pytest.fixture
    def ratelimiter(self):
        inst = hikari_test_helpers.mock_class_namespace(rate_limits.WindowedBurstRateLimiter, slots_=False)(
            __name__, 3, 3
        )
        yield inst
        with contextlib.suppress(Exception):
            inst.close()

    @pytest.mark.asyncio
    async def test_drip_if_not_throttled_and_not_ratelimited(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.remaining = 10
        ratelimiter.throttle_task = None
        ratelimiter.is_rate_limited = mock.Mock(return_value=False)
        event_loop.create_future = mock.Mock()

        await ratelimiter.acquire()

        assert ratelimiter.remaining == 9
        event_loop.create_future.assert_not_called()

    @pytest.mark.asyncio
    async def test_no_drip_if_throttle_task_is_not_None(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.remaining = 10
        ratelimiter.throttle_task = asyncio.get_running_loop().create_future()
        ratelimiter.is_rate_limited = mock.Mock(return_value=False)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        await ratelimiter.acquire()

        assert ratelimiter.remaining == 10

    @pytest.mark.asyncio
    async def test_no_drip_if_rate_limited(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.remaining = 10
        ratelimiter.throttle_task = False
        ratelimiter.is_rate_limited = mock.Mock(return_value=True)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        await ratelimiter.acquire()

        assert ratelimiter.remaining == 10

    @pytest.mark.asyncio
    async def test_task_scheduled_if_rate_limited_and_throttle_task_is_None(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.throttle_task = None
        ratelimiter.throttle = mock.AsyncMock()
        ratelimiter.is_rate_limited = mock.Mock(return_value=True)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        await ratelimiter.acquire()
        assert ratelimiter.throttle_task is not None

        ratelimiter.throttle.assert_called()

    @pytest.mark.asyncio
    async def test_task_not_scheduled_if_rate_limited_and_throttle_task_not_None(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.throttle_task = event_loop.create_future()
        old_task = ratelimiter.throttle_task
        ratelimiter.is_rate_limited = mock.Mock(return_value=True)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        await ratelimiter.acquire()
        assert old_task is ratelimiter.throttle_task, "task was rescheduled, that shouldn't happen :("

    @pytest.mark.asyncio
    async def test_future_is_added_to_queue_if_throttle_task_is_not_None(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.throttle_task = asyncio.get_running_loop().create_future()
        ratelimiter.is_rate_limited = mock.Mock(return_value=False)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        await ratelimiter.acquire()

        # use slice to prevent aborting test with index error rather than assertion error if this fails.
        assert ratelimiter.queue[-1:] == [future]

    @pytest.mark.asyncio
    async def test_future_is_added_to_queue_if_rate_limited(self, ratelimiter):
        event_loop = asyncio.get_running_loop()

        ratelimiter.throttle_task = None
        ratelimiter.is_rate_limited = mock.Mock(return_value=True)
        future = MockFuture()
        event_loop.create_future = mock.Mock(return_value=future)

        try:
            await ratelimiter.acquire()
            # use slice to prevent aborting test with index error rather than assertion error if this fails.
            assert ratelimiter.queue[-1:] == [future]
        finally:
            ratelimiter.throttle_task.cancel()

    @pytest.mark.asyncio
    async def test_throttle_consumes_queue(self):
        event_loop = asyncio.get_running_loop()

        with mock.patch.object(asyncio, "sleep"):
            with rate_limits.WindowedBurstRateLimiter(__name__, 0.001, 1) as rl:
                rl.queue = [event_loop.create_future() for _ in range(15)]
                old_queue = list(rl.queue)
                await rl.throttle()

        assert len(rl.queue) == 0
        for i, future in enumerate(old_queue):
            assert future.done(), f"future {i} was incomplete!"

    @pytest.mark.asyncio
    async def test_throttle_when_limited_sleeps_then_bursts_repeatedly(self):
        event_loop = asyncio.get_running_loop()

        window = 5
        loop_count = 0
        futures = [event_loop.create_future() for _ in range(20)]
        reset_time_iter = iter(range(int(len(futures) / window)))

        def mock_get_time_until_increase(_self, _):
            nonlocal loop_count

            for i, future in enumerate(futures):
                if i >= (window * loop_count):
                    assert not future.done(), f"future {i} was complete, expected it to be incomplete!"
                else:
                    assert future.done(), f"future {i} was incomplete, expected it to be completed!"

            loop_count += 1

            rl.remaining = window

            return next(reset_time_iter)

        stack = contextlib.ExitStack()
        rl = stack.enter_context(rate_limits.WindowedBurstRateLimiter(__name__, 0, window))
        stack.enter_context(
            mock.patch.object(
                rate_limits.WindowedBurstRateLimiter, "get_time_until_increase", new=mock_get_time_until_increase
            )
        )
        stack.enter_context(mock.patch.object(asyncio, "sleep"))

        with stack:
            rl.queue = list(futures)
            rl.increase_at = time.time()
            await rl.throttle()
            # die if we take too long...
            await asyncio.wait(futures, timeout=3)

        assert loop_count == 4
        assert len(rl.queue) == 0
        for i, future in enumerate(futures):
            assert future.done(), f"future {i} was incomplete!"

    @pytest.mark.asyncio
    async def test_throttle_resets_throttle_task(self):
        event_loop = asyncio.get_running_loop()

        with rate_limits.WindowedBurstRateLimiter(__name__, 0.001, 1) as rl:
            rl.queue = [event_loop.create_future() for _ in range(15)]
            rl.throttle_task = None
            await rl.throttle()
        assert rl.throttle_task is None

    def test_get_time_until_increase_if_not_rate_limited(self):
        with hikari_test_helpers.mock_class_namespace(rate_limits.WindowedBurstRateLimiter, slots_=False)(
            __name__, 0.01, 1
        ) as rl:
            rl.is_rate_limited = mock.Mock(return_value=False)
            assert rl.get_time_until_increase(420) == 0.0

    def test_get_time_until_increase_if_rate_limited(self):
        with hikari_test_helpers.mock_class_namespace(rate_limits.WindowedBurstRateLimiter, slots_=False)(
            __name__, 0.01, 1
        ) as rl:
            rl.is_rate_limited = mock.Mock(return_value=True)
            rl.increase_at = 420.4
            assert rl.get_time_until_increase(69.8) == 420.4 - 69.8

    def test_is_rate_limited_when_rate_limit_expired(self):
        with rate_limits.WindowedBurstRateLimiter(__name__, 27, 27) as rl:
            now = 180
            rl.increase_at = 80
            rl.remaining = 3

            assert not rl.is_rate_limited(now)

            assert rl.increase_at == 207
            assert rl.remaining == 27

    @pytest.mark.parametrize("remaining", [-1, 0, 1])
    def test_is_rate_limited_when_rate_limit_not_expired(self, remaining):
        with rate_limits.WindowedBurstRateLimiter(__name__, 403, 27) as rl:
            now = 420
            rl.increase_at = now + 69
            rl.remaining = remaining
            assert rl.is_rate_limited(now) is (remaining <= 0)


class TestExponentialBackOff:
    def test___init___raises_on_too_large_int_base(self):
        base = int(sys.float_info.max) + int(sys.float_info.max * 1 / 100)
        with pytest.raises(ValueError, match="int too large to be represented as a float"):
            rate_limits.ExponentialBackOff(base=base)

    def test___init___raises_on_too_large_int_maximum(self):
        maximum = int(sys.float_info.max) + int(sys.float_info.max * 1 / 200)
        with pytest.raises(ValueError, match="int too large to be represented as a float"):
            rate_limits.ExponentialBackOff(maximum=maximum)

    def test___init___raises_on_too_large_int_jitter_multiplier(self):
        jitter_multiplier = int(sys.float_info.max) + int(sys.float_info.max * 1 / 300)
        with pytest.raises(ValueError, match="int too large to be represented as a float"):
            rate_limits.ExponentialBackOff(jitter_multiplier=jitter_multiplier)

    def test___init___raises_on_not_finite_base(self):
        with pytest.raises(ValueError, match="base must be a finite number"):
            rate_limits.ExponentialBackOff(base=float("inf"))

    def test___init___raises_on_not_finite_maximum(self):
        with pytest.raises(ValueError, match="maximum must be a finite number"):
            rate_limits.ExponentialBackOff(maximum=float("nan"))

    def test___init___raises_on_not_finite_jitter_multiplier(self):
        with pytest.raises(ValueError, match="jitter_multiplier must be a finite number"):
            rate_limits.ExponentialBackOff(jitter_multiplier=float("inf"))

    def test_reset(self):
        eb = rate_limits.ExponentialBackOff()
        eb.increment = 10
        eb.reset()
        assert eb.increment == 0

    @pytest.mark.parametrize(("iteration", "backoff"), enumerate((1, 2, 4, 8, 16, 32)))
    def test_increment_linear(self, iteration, backoff):
        eb = rate_limits.ExponentialBackOff(2, 64, 0)

        for _ in range(iteration):
            next(eb)

        assert next(eb) == backoff

    def test_increment_raises_on_numerical_limitation(self):
        power = math.log(sys.float_info.max, 5) + 0.5
        eb = rate_limits.ExponentialBackOff(
            base=5, maximum=sys.float_info.max, jitter_multiplier=0.0, initial_increment=power
        )

        assert next(eb) == sys.float_info.max

    def test_increment_maximum(self):
        max_bound = 64
        eb = rate_limits.ExponentialBackOff(2, max_bound, 0)
        iterations = math.ceil(math.log2(max_bound))
        for _ in range(iterations):
            next(eb)

        assert next(eb) == max_bound

    def test_increment_does_not_increment_when_on_maximum(self):
        eb = rate_limits.ExponentialBackOff(2, 32, initial_increment=5, jitter_multiplier=0)

        assert eb.increment == 5

        assert next(eb) == 32

        assert eb.increment == 5

    @pytest.mark.parametrize(("iteration", "backoff"), enumerate((1, 2, 4, 8, 16, 32)))
    def test_increment_jitter(self, iteration, backoff):
        abs_tol = 1
        eb = rate_limits.ExponentialBackOff(2, 64, abs_tol)

        for _ in range(iteration):
            next(eb)

        assert math.isclose(next(eb), backoff, abs_tol=abs_tol)

    def test_iter_returns_self(self):
        eb = rate_limits.ExponentialBackOff(2, 64, 123)
        assert iter(eb) is eb



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Events that fire for auto-moderation related changes."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "AutoModActionExecutionEvent",
    "AutoModEvent",
    "AutoModRuleCreateEvent",
    "AutoModRuleDeleteEvent",
    "AutoModRuleUpdateEvent",
)

import abc
import typing

import attr

from hikari import intents
from hikari.events import base_events
from hikari.events import shard_events
from hikari.internal import attrs_extensions
from hikari.internal import typing_extensions

if typing.TYPE_CHECKING:
    from hikari import auto_mod
    from hikari import snowflakes
    from hikari import traits
    from hikari.api import shard as gateway_shard


@base_events.requires_intents(intents.Intents.AUTO_MODERATION_CONFIGURATION, intents.Intents.AUTO_MODERATION_EXECUTION)
class AutoModEvent(shard_events.ShardEvent, abc.ABC):
    """Base class for auto-moderation gateway events."""

    __slots__: typing.Sequence[str] = ()


@base_events.requires_intents(intents.Intents.AUTO_MODERATION_CONFIGURATION)
@attrs_extensions.with_copy
@attr.define(kw_only=True, weakref_slot=False)
class AutoModRuleCreateEvent(AutoModEvent):
    """Event that's fired when an auto-moderation rule is created."""

    shard: gateway_shard.GatewayShard = attr.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from ShardEvent>>.

    rule: auto_mod.AutoModRule = attr.field()
    """Object of the auto-moderation rule which was created."""

    @property
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        # <<inherited docstring from BaseEvent>>.
        return self.rule.app


@base_events.requires_intents(intents.Intents.AUTO_MODERATION_CONFIGURATION)
@attrs_extensions.with_copy
@attr.define(kw_only=True, weakref_slot=False)
class AutoModRuleUpdateEvent(AutoModEvent):
    """Event that's fired when an auto-moderation rule is updated."""

    shard: gateway_shard.GatewayShard = attr.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from ShardEvent>>.

    rule: auto_mod.AutoModRule = attr.field()
    """Object of the auto-moderation rule which was updated."""

    @property
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        # <<inherited docstring from BaseEvent>>.
        return self.rule.app


@base_events.requires_intents(intents.Intents.AUTO_MODERATION_CONFIGURATION)
@attrs_extensions.with_copy
@attr.define(kw_only=True, weakref_slot=False)
class AutoModRuleDeleteEvent(AutoModEvent):
    """Event that's fired when an auto-moderation rule is deleted."""

    shard: gateway_shard.GatewayShard = attr.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from ShardEvent>>.

    rule: auto_mod.AutoModRule = attr.field()
    """Object of the auto-moderation rule which was deleted."""

    @property
    @typing_extensions.override
    def app(self) -> traits.RESTAware:
        # <<inherited docstring from BaseEvent>>.
        return self.rule.app


@base_events.requires_intents(intents.Intents.AUTO_MODERATION_EXECUTION)
@attrs_extensions.with_copy
@attr.define(kw_only=True, weakref_slot=False)
class AutoModActionExecutionEvent(AutoModEvent):
    """Event that's fired when an auto-mod action is executed."""

    app: traits.RESTAware = attr.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from Event>>.

    shard: gateway_shard.GatewayShard = attr.field(metadata={attrs_extensions.SKIP_DEEP_COPY: True})
    # <<inherited docstring from ShardEvent>>.

    guild_id: snowflakes.Snowflake = attr.field(repr=True)
    """ID of the guild this action was executed in."""

    action: auto_mod.PartialAutoModAction = attr.field(repr=False)
    """Object of the action which was executed."""

    rule_id: snowflakes.Snowflake = attr.field()
    """ID of the rule which was triggered."""

    rule_trigger_type: int | auto_mod.AutoModTriggerType | None = attr.field(repr=False)
    """Type of the rule which was triggered."""

    user_id: snowflakes.Snowflake = attr.field(repr=False)
    """ID of the user who generated the context which triggered this."""

    channel_id: snowflakes.Snowflake | None = attr.field(repr=False)
    """ID of the channel the matching context was sent to.

    This will be [`None`][] if the message was blocked by auto-moderation
    of the matched content wasn't in a channel.
    """

    message_id: snowflakes.Snowflake | None = attr.field(repr=False)
    """ID of the message the matching context was sent in.

    This will be [`None`][] if the message was blocked by auto-moderation
    or the matched content wasn't in a message.
    """

    alert_system_message_id: snowflakes.Snowflake | None = attr.field(repr=False)
    """ID of any system auto-moderation messages posted as a result of this action.

    This will only be provided for `SEND_ALERT_MESSAGE` actions.
    """

    content: str | None = attr.field(repr=False)
    """The user generated content which matched this rule.

    This will only be provided if the `MESSAGE_CONTENT` intent has
    been declared.
    """

    matched_keyword: str | None = attr.field(repr=False)
    """The word or phrase configured in the rule which was triggered, if it's a keyword trigger."""

    matched_content: str | None = attr.field(repr=False)
    """The substring in content which triggered the rule.

    This will only be provided if the `MESSAGE_CONTENT` intent has
    been declared and this is a keyword or keyword preset trigger.
    """



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Documentation pages generation."""

from __future__ import annotations

import os

from pipelines import config
from pipelines import nox


def _setup_environ(session: nox.Session) -> None:
    if "--no-refs" in session.posargs:
        session.env["ENABLE_MKDOCSTRINGS"] = "false"

    if os.environ.get("READTHEDOCS_VERSION_NAME", "master") == "master":
        session.env["HIDE_EMPTY_TOWNCRIER_DRAFT"] = "false"


@nox.session()
def mkdocs(session: nox.Session) -> None:
    """Generate docs using mkdocs."""
    _setup_environ(session)

    nox.sync(session, self=True, groups=["mkdocs", "ruff"])

    session.run("mkdocs", "build", "-d", config.DOCUMENTATION_OUTPUT_PATH)


@nox.session()
def mkdocs_serve(session: nox.Session) -> None:
    """Start an HTTP server that serves the generated docs in real time."""
    _setup_environ(session)

    nox.sync(session, self=True, groups=["mkdocs", "ruff"])

    if "--no-reload" in session.posargs:
        session.run("mkdocs", "serve", "--no-livereload")
    else:
        session.run("mkdocs", "serve")



# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Models and enums used for Discord's Modals interaction flow."""

from __future__ import annotations

__all__: typing.Sequence[str] = (
    "ModalInteraction",
    "ModalInteraction",
    "ModalInteractionMetadata",
    "ModalResponseTypesT",
)

import typing

import attrs

from hikari.interactions import base_interactions
from hikari.internal import attrs_extensions

if typing.TYPE_CHECKING:
    from hikari import components as components_
    from hikari import messages
    from hikari import snowflakes
    from hikari.api import special_endpoints

ModalResponseTypesT = typing.Literal[
    base_interactions.ResponseType.MESSAGE_CREATE,
    4,
    base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE,
    5,
    base_interactions.ResponseType.MESSAGE_UPDATE,
    7,
    base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE,
    6,
]
"""Type-hint of the response types which are valid for a modal interaction.

The following types are valid for this:

* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_CREATE`][]/`4`
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE`][]/`5`
* [`hikari.interactions.base_interactions.ResponseType.MESSAGE_UPDATE`][]/`7`
* [`hikari.interactions.base_interactions.ResponseType.DEFERRED_MESSAGE_UPDATE`][]/`6`
"""


@attrs_extensions.with_copy
@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ModalInteraction(
    base_interactions.MessageResponseMixin[ModalResponseTypesT], base_interactions.PremiumResponseMixin
):
    """Represents a modal interaction on Discord."""

    custom_id: str = attrs.field(eq=False, hash=False, repr=True)
    """The custom id of the modal."""

    message: messages.Message | None = attrs.field(eq=False, repr=False)
    """The message whose component triggered the modal.

    This will be [`None`][] if the modal was a response to a command.
    """

    components: typing.Sequence[components_.ModalActionRowComponent] = attrs.field(eq=False, hash=False, repr=True)
    """Components in the modal."""

    def build_response(self) -> special_endpoints.InteractionMessageBuilder:
        """Get a message response builder for use in the REST server flow.

        !!! note
            For interactions received over the gateway
            [`hikari.interactions.modal_interactions.ModalInteraction.create_initial_response`][]
            should be used to set the interaction response message.

        Examples
        --------
        ```py
        async def handle_modal_interaction(
            interaction: ModalInteraction,
        ) -> InteractionMessageBuilder:
            return (
                interaction.build_response()
                .add_embed(Embed(description="Hi there"))
                .set_content("Konnichiwa")
            )
        ```

        Returns
        -------
        hikari.api.special_endpoints.InteractionMessageBuilder
            Interaction message response builder object.
        """
        return self.app.rest.interaction_message_builder(base_interactions.ResponseType.MESSAGE_CREATE)

    def build_deferred_response(self) -> special_endpoints.InteractionDeferredBuilder:
        """Get a deferred message response builder for use in the REST server flow.

        !!! note
            For interactions received over the gateway
            [`hikari.interactions.modal_interactions.ModalInteraction.create_initial_response`][]
            should be used to set the interaction response message.

        !!! note
            Unlike [`hikari.api.special_endpoints.InteractionMessageBuilder`][],
            the result of this call can be returned as is without any modifications
            being made to it.

        Returns
        -------
        hikari.api.special_endpoints.InteractionDeferredBuilder
            Deferred interaction message response builder object.
        """
        return self.app.rest.interaction_deferred_builder(base_interactions.ResponseType.DEFERRED_MESSAGE_CREATE)


@attrs.define(unsafe_hash=True, kw_only=True, weakref_slot=False)
class ModalInteractionMetadata(base_interactions.PartialInteractionMetadata):
    """The interaction metadata for a modal initiated message."""

    original_response_message_id: snowflakes.Snowflake | None = attrs.field(eq=False, hash=False, repr=True)
    """The ID of the original response message, present only on follow-up messages."""

    triggering_interaction_metadata: base_interactions.PartialInteractionMetadata = attrs.field(
        eq=False, hash=False, repr=True
    )
    """The metadata for the interaction that was used to open the modal."""
